import unittest

from xdrone.shared.boundary_config import BoundaryConfig
from xdrone.shared.safety_check_error import SafetyCheckError
from xdrone.shared.state import State


class BoundaryConfigTest(unittest.TestCase):
    def test_init_with_wrong_parameter_should_give_error(self):
        with self.assertRaises(ValueError) as context:
            BoundaryConfig(max_seconds=-1)
        self.assertTrue("max_seconds should >= 0" in str(context.exception))

        with self.assertRaises(ValueError) as context:
            BoundaryConfig(max_x_meters=-1, min_x_meters=1)
        self.assertTrue("max_x_meters should >= min_x_meters" in str(context.exception))

        with self.assertRaises(ValueError) as context:
            BoundaryConfig(max_y_meters=-1, min_y_meters=1)
        self.assertTrue("max_y_meters should >= min_y_meters" in str(context.exception))

        with self.assertRaises(ValueError) as context:
            BoundaryConfig(max_z_meters=-1, min_z_meters=1)
        self.assertTrue("max_z_meters should >= min_z_meters" in str(context.exception))

    def test_init_values(self):
        boundary_config = BoundaryConfig()
        self.assertEqual(float("inf"), boundary_config._max_seconds)
        self.assertEqual(float("inf"), boundary_config._max_x_meters)
        self.assertEqual(float("inf"), boundary_config._max_y_meters)
        self.assertEqual(float("inf"), boundary_config._max_z_meters)
        self.assertEqual(float("-inf"), boundary_config._min_x_meters)
        self.assertEqual(float("-inf"), boundary_config._min_y_meters)
        self.assertEqual(float("-inf"), boundary_config._min_z_meters)

    def test_no_limit_values(self):
        boundary_config = BoundaryConfig.no_limit()
        self.assertEqual(float("inf"), boundary_config._max_seconds)
        self.assertEqual(float("inf"), boundary_config._max_x_meters)
        self.assertEqual(float("inf"), boundary_config._max_y_meters)
        self.assertEqual(float("inf"), boundary_config._max_z_meters)
        self.assertEqual(float("-inf"), boundary_config._min_x_meters)
        self.assertEqual(float("-inf"), boundary_config._min_y_meters)
        self.assertEqual(float("-inf"), boundary_config._min_z_meters)

    def test_str(self):
        boundary_config = BoundaryConfig(max_seconds=1, max_x_meters=2, max_y_meters=3, max_z_meters=4,
                                         min_x_meters=-2, min_y_meters=-3, min_z_meters=-4)
        self.assertEqual("BoundaryConfig: { max_seconds: 1, x_range_meters: (-2, 2), " +
                         "y_range_meters: (-3, 3), z_range_meters: (-4, 4) }",
                         str(boundary_config))

    def test_eq(self):
        self.assertEqual(BoundaryConfig(), BoundaryConfig())
        self.assertNotEqual(BoundaryConfig(max_seconds=0), BoundaryConfig(max_seconds=1))
        self.assertNotEqual(BoundaryConfig(max_x_meters=0), BoundaryConfig(max_x_meters=1))
        self.assertNotEqual(BoundaryConfig(max_y_meters=0), BoundaryConfig(max_y_meters=1))
        self.assertNotEqual(BoundaryConfig(max_z_meters=0), BoundaryConfig(max_z_meters=1))
        self.assertNotEqual(BoundaryConfig(min_x_meters=0), BoundaryConfig(min_x_meters=-1))
        self.assertNotEqual(BoundaryConfig(min_y_meters=0), BoundaryConfig(min_y_meters=-1))
        self.assertNotEqual(BoundaryConfig(min_z_meters=0), BoundaryConfig(min_z_meters=-1))
        self.assertNotEqual(None, BoundaryConfig())

    def test_check_state_if_beyond_limit_should_give_error(self):
        boundary_config = BoundaryConfig(max_seconds=0, max_x_meters=0, max_y_meters=0, max_z_meters=0,
                                         min_x_meters=0, min_y_meters=0, min_z_meters=0)
        with self.assertRaises(SafetyCheckError) as context:
            boundary_config.check_state('default', State(x_meters=10))
        self.assertTrue("Drone 'default': the x coordinate 10 will go beyond its upper limit 0"
                        in str(context.exception))

        with self.assertRaises(SafetyCheckError) as context:
            boundary_config.check_state('default', State(y_meters=10))
        self.assertTrue("Drone 'default': the y coordinate 10 will go beyond its upper limit 0"
                        in str(context.exception))

        with self.assertRaises(SafetyCheckError) as context:
            boundary_config.check_state('default', State(z_meters=10))
        self.assertTrue("Drone 'default': the z coordinate 10 will go beyond its upper limit 0"
                        in str(context.exception))

        with self.assertRaises(SafetyCheckError) as context:
            boundary_config.check_state('default', State(x_meters=-10))
        self.assertTrue("Drone 'default': the x coordinate -10 will go beyond its lower limit 0"
                        in str(context.exception))

        with self.assertRaises(SafetyCheckError) as context:
            boundary_config.check_state('default', State(y_meters=-10))
        self.assertTrue("Drone 'default': the y coordinate -10 will go beyond its lower limit 0"
                        in str(context.exception))

        with self.assertRaises(SafetyCheckError) as context:
            boundary_config.check_state('default', State(z_meters=-10))
        self.assertTrue("Drone 'default': the z coordinate -10 will go beyond its lower limit 0"
                        in str(context.exception))

        with self.assertRaises(SafetyCheckError) as context:
            boundary_config.check_state('default', State(time_used_seconds=10))
        self.assertTrue("Drone 'default': the time used 10 seconds will go beyond the time limit 0 seconds"
                        in str(context.exception))

    def test_check_state_if_no_limit_should_not_give_error(self):
        boundary_config = BoundaryConfig()
        for number in [1e-5, 1e-3, 1e0, 1e3, 1e5, 1e10, float('inf')]:
            boundary_config.check_state('default', State(x_meters=number))
            boundary_config.check_state('default', State(y_meters=number))
            boundary_config.check_state('default', State(z_meters=number))
            boundary_config.check_state('default', State(x_meters=-number))
            boundary_config.check_state('default', State(y_meters=-number))
            boundary_config.check_state('default', State(z_meters=-number))
            boundary_config.check_state('default', State(time_used_seconds=number))



from __future__ import annotations

import copy
from math import isclose


class State:
    def __init__(self, has_taken_off: bool = False, time_used_seconds: float = 0,
                 x_meters: float = 0, y_meters: float = 0, z_meters: float = 0, orientation_degrees: float = 0):
        self._has_taken_off = has_taken_off
        self._time_used_seconds = time_used_seconds
        self._x_meters = x_meters
        self._y_meters = y_meters
        self._z_meters = z_meters
        self._orientation_degrees = orientation_degrees

    @property
    def has_taken_off(self) -> bool:
        return copy.deepcopy(self._has_taken_off)

    @property
    def time_used_seconds(self) -> float:
        return copy.deepcopy(self._time_used_seconds)

    @property
    def x_meters(self) -> float:
        return copy.deepcopy(self._x_meters)

    @property
    def y_meters(self) -> float:
        return copy.deepcopy(self._y_meters)

    @property
    def z_meters(self) -> float:
        return copy.deepcopy(self._z_meters)

    @property
    def orientation_degrees(self) -> float:
        return copy.deepcopy(self._orientation_degrees)

    def copy_and_set_has_taken_off(self, has_taken_off: bool) -> State:
        return State(has_taken_off=has_taken_off, time_used_seconds=self.time_used_seconds,
                     x_meters=self.x_meters, y_meters=self.y_meters, z_meters=self.z_meters,
                     orientation_degrees=self.orientation_degrees)

    def copy_and_set_time_used_seconds(self, time_used_seconds: float) -> State:
        return State(has_taken_off=self.has_taken_off, time_used_seconds=time_used_seconds,
                     x_meters=self.x_meters, y_meters=self.y_meters, z_meters=self.z_meters,
                     orientation_degrees=self.orientation_degrees)

    def copy_and_set_x_meters(self, x_meters: float) -> State:
        return State(has_taken_off=self.has_taken_off, time_used_seconds=self.time_used_seconds,
                     x_meters=x_meters, y_meters=self.y_meters, z_meters=self.z_meters,
                     orientation_degrees=self.orientation_degrees)

    def copy_and_set_y_meters(self, y_meters: float) -> State:
        return State(has_taken_off=self.has_taken_off, time_used_seconds=self.time_used_seconds,
                     x_meters=self.x_meters, y_meters=y_meters, z_meters=self.z_meters,
                     orientation_degrees=self.orientation_degrees)

    def copy_and_set_z_meters(self, z_meters: float) -> State:
        return State(has_taken_off=self.has_taken_off, time_used_seconds=self.time_used_seconds,
                     x_meters=self.x_meters, y_meters=self.y_meters, z_meters=z_meters,
                     orientation_degrees=self.orientation_degrees)

    def copy_and_set_orientation_degrees(self, orientation_degrees: float) -> State:
        return State(has_taken_off=self.has_taken_off, time_used_seconds=self.time_used_seconds,
                     x_meters=self.x_meters, y_meters=self.y_meters, z_meters=self.z_meters,
                     orientation_degrees=orientation_degrees)

    def __str__(self):
        return ("State: {{ has_taken_off: {}, time_used_seconds: {}, "
                .format(self.has_taken_off, self.time_used_seconds) +
                "x_meters: {}, y_meters: {}, z_meters: {}, orientation_degrees: {} }}"
                .format(self.x_meters, self.y_meters, self.z_meters, self.orientation_degrees))

    def __eq__(self, other):
        if isinstance(other, State):
            return other.has_taken_off == self.has_taken_off \
                   and isclose(other.time_used_seconds, self.time_used_seconds) \
                   and isclose(other.x_meters, self.x_meters) and isclose(other.y_meters, self.y_meters) \
                   and isclose(other.z_meters, self.z_meters) \
                   and isclose(other.orientation_degrees, self.orientation_degrees)
        return False



# Generated from xDroneParser.g4 by ANTLR 4.9.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .xDroneParser import xDroneParser
else:
    from xDroneParser import xDroneParser

# This class defines a complete listener for a parse tree produced by xDroneParser.
class xDroneParserListener(ParseTreeListener):

    # Enter a parse tree produced by xDroneParser#prog.
    def enterProg(self, ctx:xDroneParser.ProgContext):
        pass

    # Exit a parse tree produced by xDroneParser#prog.
    def exitProg(self, ctx:xDroneParser.ProgContext):
        pass


    # Enter a parse tree produced by xDroneParser#commands.
    def enterCommands(self, ctx:xDroneParser.CommandsContext):
        pass

    # Exit a parse tree produced by xDroneParser#commands.
    def exitCommands(self, ctx:xDroneParser.CommandsContext):
        pass


    # Enter a parse tree produced by xDroneParser#takeoff.
    def enterTakeoff(self, ctx:xDroneParser.TakeoffContext):
        pass

    # Exit a parse tree produced by xDroneParser#takeoff.
    def exitTakeoff(self, ctx:xDroneParser.TakeoffContext):
        pass


    # Enter a parse tree produced by xDroneParser#land.
    def enterLand(self, ctx:xDroneParser.LandContext):
        pass

    # Exit a parse tree produced by xDroneParser#land.
    def exitLand(self, ctx:xDroneParser.LandContext):
        pass


    # Enter a parse tree produced by xDroneParser#up.
    def enterUp(self, ctx:xDroneParser.UpContext):
        pass

    # Exit a parse tree produced by xDroneParser#up.
    def exitUp(self, ctx:xDroneParser.UpContext):
        pass


    # Enter a parse tree produced by xDroneParser#down.
    def enterDown(self, ctx:xDroneParser.DownContext):
        pass

    # Exit a parse tree produced by xDroneParser#down.
    def exitDown(self, ctx:xDroneParser.DownContext):
        pass


    # Enter a parse tree produced by xDroneParser#left.
    def enterLeft(self, ctx:xDroneParser.LeftContext):
        pass

    # Exit a parse tree produced by xDroneParser#left.
    def exitLeft(self, ctx:xDroneParser.LeftContext):
        pass


    # Enter a parse tree produced by xDroneParser#right.
    def enterRight(self, ctx:xDroneParser.RightContext):
        pass

    # Exit a parse tree produced by xDroneParser#right.
    def exitRight(self, ctx:xDroneParser.RightContext):
        pass


    # Enter a parse tree produced by xDroneParser#forward.
    def enterForward(self, ctx:xDroneParser.ForwardContext):
        pass

    # Exit a parse tree produced by xDroneParser#forward.
    def exitForward(self, ctx:xDroneParser.ForwardContext):
        pass


    # Enter a parse tree produced by xDroneParser#backward.
    def enterBackward(self, ctx:xDroneParser.BackwardContext):
        pass

    # Exit a parse tree produced by xDroneParser#backward.
    def exitBackward(self, ctx:xDroneParser.BackwardContext):
        pass


    # Enter a parse tree produced by xDroneParser#rotateLeft.
    def enterRotateLeft(self, ctx:xDroneParser.RotateLeftContext):
        pass

    # Exit a parse tree produced by xDroneParser#rotateLeft.
    def exitRotateLeft(self, ctx:xDroneParser.RotateLeftContext):
        pass


    # Enter a parse tree produced by xDroneParser#rotateRight.
    def enterRotateRight(self, ctx:xDroneParser.RotateRightContext):
        pass

    # Exit a parse tree produced by xDroneParser#rotateRight.
    def exitRotateRight(self, ctx:xDroneParser.RotateRightContext):
        pass


    # Enter a parse tree produced by xDroneParser#wait.
    def enterWait(self, ctx:xDroneParser.WaitContext):
        pass

    # Exit a parse tree produced by xDroneParser#wait.
    def exitWait(self, ctx:xDroneParser.WaitContext):
        pass


    # Enter a parse tree produced by xDroneParser#declare.
    def enterDeclare(self, ctx:xDroneParser.DeclareContext):
        pass

    # Exit a parse tree produced by xDroneParser#declare.
    def exitDeclare(self, ctx:xDroneParser.DeclareContext):
        pass


    # Enter a parse tree produced by xDroneParser#declareAssign.
    def enterDeclareAssign(self, ctx:xDroneParser.DeclareAssignContext):
        pass

    # Exit a parse tree produced by xDroneParser#declareAssign.
    def exitDeclareAssign(self, ctx:xDroneParser.DeclareAssignContext):
        pass


    # Enter a parse tree produced by xDroneParser#assignVectorElem.
    def enterAssignVectorElem(self, ctx:xDroneParser.AssignVectorElemContext):
        pass

    # Exit a parse tree produced by xDroneParser#assignVectorElem.
    def exitAssignVectorElem(self, ctx:xDroneParser.AssignVectorElemContext):
        pass


    # Enter a parse tree produced by xDroneParser#assignListElem.
    def enterAssignListElem(self, ctx:xDroneParser.AssignListElemContext):
        pass

    # Exit a parse tree produced by xDroneParser#assignListElem.
    def exitAssignListElem(self, ctx:xDroneParser.AssignListElemContext):
        pass


    # Enter a parse tree produced by xDroneParser#assignIdent.
    def enterAssignIdent(self, ctx:xDroneParser.AssignIdentContext):
        pass

    # Exit a parse tree produced by xDroneParser#assignIdent.
    def exitAssignIdent(self, ctx:xDroneParser.AssignIdentContext):
        pass


    # Enter a parse tree produced by xDroneParser#del.
    def enterDel(self, ctx:xDroneParser.DelContext):
        pass

    # Exit a parse tree produced by xDroneParser#del.
    def exitDel(self, ctx:xDroneParser.DelContext):
        pass


    # Enter a parse tree produced by xDroneParser#insert.
    def enterInsert(self, ctx:xDroneParser.InsertContext):
        pass

    # Exit a parse tree produced by xDroneParser#insert.
    def exitInsert(self, ctx:xDroneParser.InsertContext):
        pass


    # Enter a parse tree produced by xDroneParser#remove.
    def enterRemove(self, ctx:xDroneParser.RemoveContext):
        pass

    # Exit a parse tree produced by xDroneParser#remove.
    def exitRemove(self, ctx:xDroneParser.RemoveContext):
        pass


    # Enter a parse tree produced by xDroneParser#procedureCall.
    def enterProcedureCall(self, ctx:xDroneParser.ProcedureCallContext):
        pass

    # Exit a parse tree produced by xDroneParser#procedureCall.
    def exitProcedureCall(self, ctx:xDroneParser.ProcedureCallContext):
        pass


    # Enter a parse tree produced by xDroneParser#if.
    def enterIf(self, ctx:xDroneParser.IfContext):
        pass

    # Exit a parse tree produced by xDroneParser#if.
    def exitIf(self, ctx:xDroneParser.IfContext):
        pass


    # Enter a parse tree produced by xDroneParser#while.
    def enterWhile(self, ctx:xDroneParser.WhileContext):
        pass

    # Exit a parse tree produced by xDroneParser#while.
    def exitWhile(self, ctx:xDroneParser.WhileContext):
        pass


    # Enter a parse tree produced by xDroneParser#for.
    def enterFor(self, ctx:xDroneParser.ForContext):
        pass

    # Exit a parse tree produced by xDroneParser#for.
    def exitFor(self, ctx:xDroneParser.ForContext):
        pass


    # Enter a parse tree produced by xDroneParser#repeat.
    def enterRepeat(self, ctx:xDroneParser.RepeatContext):
        pass

    # Exit a parse tree produced by xDroneParser#repeat.
    def exitRepeat(self, ctx:xDroneParser.RepeatContext):
        pass


    # Enter a parse tree produced by xDroneParser#return.
    def enterReturn(self, ctx:xDroneParser.ReturnContext):
        pass

    # Exit a parse tree produced by xDroneParser#return.
    def exitReturn(self, ctx:xDroneParser.ReturnContext):
        pass


    # Enter a parse tree produced by xDroneParser#parallel.
    def enterParallel(self, ctx:xDroneParser.ParallelContext):
        pass

    # Exit a parse tree produced by xDroneParser#parallel.
    def exitParallel(self, ctx:xDroneParser.ParallelContext):
        pass


    # Enter a parse tree produced by xDroneParser#ident.
    def enterIdent(self, ctx:xDroneParser.IdentContext):
        pass

    # Exit a parse tree produced by xDroneParser#ident.
    def exitIdent(self, ctx:xDroneParser.IdentContext):
        pass


    # Enter a parse tree produced by xDroneParser#listElem.
    def enterListElem(self, ctx:xDroneParser.ListElemContext):
        pass

    # Exit a parse tree produced by xDroneParser#listElem.
    def exitListElem(self, ctx:xDroneParser.ListElemContext):
        pass


    # Enter a parse tree produced by xDroneParser#vectorX.
    def enterVectorX(self, ctx:xDroneParser.VectorXContext):
        pass

    # Exit a parse tree produced by xDroneParser#vectorX.
    def exitVectorX(self, ctx:xDroneParser.VectorXContext):
        pass


    # Enter a parse tree produced by xDroneParser#vectorY.
    def enterVectorY(self, ctx:xDroneParser.VectorYContext):
        pass

    # Exit a parse tree produced by xDroneParser#vectorY.
    def exitVectorY(self, ctx:xDroneParser.VectorYContext):
        pass


    # Enter a parse tree produced by xDroneParser#vectorZ.
    def enterVectorZ(self, ctx:xDroneParser.VectorZContext):
        pass

    # Exit a parse tree produced by xDroneParser#vectorZ.
    def exitVectorZ(self, ctx:xDroneParser.VectorZContext):
        pass


    # Enter a parse tree produced by xDroneParser#funcIdent.
    def enterFuncIdent(self, ctx:xDroneParser.FuncIdentContext):
        pass

    # Exit a parse tree produced by xDroneParser#funcIdent.
    def exitFuncIdent(self, ctx:xDroneParser.FuncIdentContext):
        pass


    # Enter a parse tree produced by xDroneParser#call.
    def enterCall(self, ctx:xDroneParser.CallContext):
        pass

    # Exit a parse tree produced by xDroneParser#call.
    def exitCall(self, ctx:xDroneParser.CallContext):
        pass


    # Enter a parse tree produced by xDroneParser#argList.
    def enterArgList(self, ctx:xDroneParser.ArgListContext):
        pass

    # Exit a parse tree produced by xDroneParser#argList.
    def exitArgList(self, ctx:xDroneParser.ArgListContext):
        pass


    # Enter a parse tree produced by xDroneParser#function.
    def enterFunction(self, ctx:xDroneParser.FunctionContext):
        pass

    # Exit a parse tree produced by xDroneParser#function.
    def exitFunction(self, ctx:xDroneParser.FunctionContext):
        pass


    # Enter a parse tree produced by xDroneParser#procedure.
    def enterProcedure(self, ctx:xDroneParser.ProcedureContext):
        pass

    # Exit a parse tree produced by xDroneParser#procedure.
    def exitProcedure(self, ctx:xDroneParser.ProcedureContext):
        pass


    # Enter a parse tree produced by xDroneParser#paramList.
    def enterParamList(self, ctx:xDroneParser.ParamListContext):
        pass

    # Exit a parse tree produced by xDroneParser#paramList.
    def exitParamList(self, ctx:xDroneParser.ParamListContext):
        pass


    # Enter a parse tree produced by xDroneParser#intType.
    def enterIntType(self, ctx:xDroneParser.IntTypeContext):
        pass

    # Exit a parse tree produced by xDroneParser#intType.
    def exitIntType(self, ctx:xDroneParser.IntTypeContext):
        pass


    # Enter a parse tree produced by xDroneParser#decimalType.
    def enterDecimalType(self, ctx:xDroneParser.DecimalTypeContext):
        pass

    # Exit a parse tree produced by xDroneParser#decimalType.
    def exitDecimalType(self, ctx:xDroneParser.DecimalTypeContext):
        pass


    # Enter a parse tree produced by xDroneParser#stringType.
    def enterStringType(self, ctx:xDroneParser.StringTypeContext):
        pass

    # Exit a parse tree produced by xDroneParser#stringType.
    def exitStringType(self, ctx:xDroneParser.StringTypeContext):
        pass


    # Enter a parse tree produced by xDroneParser#booleanType.
    def enterBooleanType(self, ctx:xDroneParser.BooleanTypeContext):
        pass

    # Exit a parse tree produced by xDroneParser#booleanType.
    def exitBooleanType(self, ctx:xDroneParser.BooleanTypeContext):
        pass


    # Enter a parse tree produced by xDroneParser#vectorType.
    def enterVectorType(self, ctx:xDroneParser.VectorTypeContext):
        pass

    # Exit a parse tree produced by xDroneParser#vectorType.
    def exitVectorType(self, ctx:xDroneParser.VectorTypeContext):
        pass


    # Enter a parse tree produced by xDroneParser#droneType.
    def enterDroneType(self, ctx:xDroneParser.DroneTypeContext):
        pass

    # Exit a parse tree produced by xDroneParser#droneType.
    def exitDroneType(self, ctx:xDroneParser.DroneTypeContext):
        pass


    # Enter a parse tree produced by xDroneParser#listType.
    def enterListType(self, ctx:xDroneParser.ListTypeContext):
        pass

    # Exit a parse tree produced by xDroneParser#listType.
    def exitListType(self, ctx:xDroneParser.ListTypeContext):
        pass


    # Enter a parse tree produced by xDroneParser#parentheses.
    def enterParentheses(self, ctx:xDroneParser.ParenthesesContext):
        pass

    # Exit a parse tree produced by xDroneParser#parentheses.
    def exitParentheses(self, ctx:xDroneParser.ParenthesesContext):
        pass


    # Enter a parse tree produced by xDroneParser#plusMinus.
    def enterPlusMinus(self, ctx:xDroneParser.PlusMinusContext):
        pass

    # Exit a parse tree produced by xDroneParser#plusMinus.
    def exitPlusMinus(self, ctx:xDroneParser.PlusMinusContext):
        pass


    # Enter a parse tree produced by xDroneParser#compare.
    def enterCompare(self, ctx:xDroneParser.CompareContext):
        pass

    # Exit a parse tree produced by xDroneParser#compare.
    def exitCompare(self, ctx:xDroneParser.CompareContext):
        pass


    # Enter a parse tree produced by xDroneParser#intExpr.
    def enterIntExpr(self, ctx:xDroneParser.IntExprContext):
        pass

    # Exit a parse tree produced by xDroneParser#intExpr.
    def exitIntExpr(self, ctx:xDroneParser.IntExprContext):
        pass


    # Enter a parse tree produced by xDroneParser#or.
    def enterOr(self, ctx:xDroneParser.OrContext):
        pass

    # Exit a parse tree produced by xDroneParser#or.
    def exitOr(self, ctx:xDroneParser.OrContext):
        pass


    # Enter a parse tree produced by xDroneParser#decimalExpr.
    def enterDecimalExpr(self, ctx:xDroneParser.DecimalExprContext):
        pass

    # Exit a parse tree produced by xDroneParser#decimalExpr.
    def exitDecimalExpr(self, ctx:xDroneParser.DecimalExprContext):
        pass


    # Enter a parse tree produced by xDroneParser#trueExpr.
    def enterTrueExpr(self, ctx:xDroneParser.TrueExprContext):
        pass

    # Exit a parse tree produced by xDroneParser#trueExpr.
    def exitTrueExpr(self, ctx:xDroneParser.TrueExprContext):
        pass


    # Enter a parse tree produced by xDroneParser#listElemExpr.
    def enterListElemExpr(self, ctx:xDroneParser.ListElemExprContext):
        pass

    # Exit a parse tree produced by xDroneParser#listElemExpr.
    def exitListElemExpr(self, ctx:xDroneParser.ListElemExprContext):
        pass


    # Enter a parse tree produced by xDroneParser#positNegate.
    def enterPositNegate(self, ctx:xDroneParser.PositNegateContext):
        pass

    # Exit a parse tree produced by xDroneParser#positNegate.
    def exitPositNegate(self, ctx:xDroneParser.PositNegateContext):
        pass


    # Enter a parse tree produced by xDroneParser#concat.
    def enterConcat(self, ctx:xDroneParser.ConcatContext):
        pass

    # Exit a parse tree produced by xDroneParser#concat.
    def exitConcat(self, ctx:xDroneParser.ConcatContext):
        pass


    # Enter a parse tree produced by xDroneParser#list.
    def enterList(self, ctx:xDroneParser.ListContext):
        pass

    # Exit a parse tree produced by xDroneParser#list.
    def exitList(self, ctx:xDroneParser.ListContext):
        pass


    # Enter a parse tree produced by xDroneParser#vectorYExpr.
    def enterVectorYExpr(self, ctx:xDroneParser.VectorYExprContext):
        pass

    # Exit a parse tree produced by xDroneParser#vectorYExpr.
    def exitVectorYExpr(self, ctx:xDroneParser.VectorYExprContext):
        pass


    # Enter a parse tree produced by xDroneParser#stringExpr.
    def enterStringExpr(self, ctx:xDroneParser.StringExprContext):
        pass

    # Exit a parse tree produced by xDroneParser#stringExpr.
    def exitStringExpr(self, ctx:xDroneParser.StringExprContext):
        pass


    # Enter a parse tree produced by xDroneParser#identExpr.
    def enterIdentExpr(self, ctx:xDroneParser.IdentExprContext):
        pass

    # Exit a parse tree produced by xDroneParser#identExpr.
    def exitIdentExpr(self, ctx:xDroneParser.IdentExprContext):
        pass


    # Enter a parse tree produced by xDroneParser#not.
    def enterNot(self, ctx:xDroneParser.NotContext):
        pass

    # Exit a parse tree produced by xDroneParser#not.
    def exitNot(self, ctx:xDroneParser.NotContext):
        pass


    # Enter a parse tree produced by xDroneParser#vectorXExpr.
    def enterVectorXExpr(self, ctx:xDroneParser.VectorXExprContext):
        pass

    # Exit a parse tree produced by xDroneParser#vectorXExpr.
    def exitVectorXExpr(self, ctx:xDroneParser.VectorXExprContext):
        pass


    # Enter a parse tree produced by xDroneParser#size.
    def enterSize(self, ctx:xDroneParser.SizeContext):
        pass

    # Exit a parse tree produced by xDroneParser#size.
    def exitSize(self, ctx:xDroneParser.SizeContext):
        pass


    # Enter a parse tree produced by xDroneParser#and.
    def enterAnd(self, ctx:xDroneParser.AndContext):
        pass

    # Exit a parse tree produced by xDroneParser#and.
    def exitAnd(self, ctx:xDroneParser.AndContext):
        pass


    # Enter a parse tree produced by xDroneParser#functionCall.
    def enterFunctionCall(self, ctx:xDroneParser.FunctionCallContext):
        pass

    # Exit a parse tree produced by xDroneParser#functionCall.
    def exitFunctionCall(self, ctx:xDroneParser.FunctionCallContext):
        pass


    # Enter a parse tree produced by xDroneParser#vectorZExpr.
    def enterVectorZExpr(self, ctx:xDroneParser.VectorZExprContext):
        pass

    # Exit a parse tree produced by xDroneParser#vectorZExpr.
    def exitVectorZExpr(self, ctx:xDroneParser.VectorZExprContext):
        pass


    # Enter a parse tree produced by xDroneParser#vector.
    def enterVector(self, ctx:xDroneParser.VectorContext):
        pass

    # Exit a parse tree produced by xDroneParser#vector.
    def exitVector(self, ctx:xDroneParser.VectorContext):
        pass


    # Enter a parse tree produced by xDroneParser#falseExpr.
    def enterFalseExpr(self, ctx:xDroneParser.FalseExprContext):
        pass

    # Exit a parse tree produced by xDroneParser#falseExpr.
    def exitFalseExpr(self, ctx:xDroneParser.FalseExprContext):
        pass


    # Enter a parse tree produced by xDroneParser#multiDivide.
    def enterMultiDivide(self, ctx:xDroneParser.MultiDivideContext):
        pass

    # Exit a parse tree produced by xDroneParser#multiDivide.
    def exitMultiDivide(self, ctx:xDroneParser.MultiDivideContext):
        pass


    # Enter a parse tree produced by xDroneParser#equality.
    def enterEquality(self, ctx:xDroneParser.EqualityContext):
        pass

    # Exit a parse tree produced by xDroneParser#equality.
    def exitEquality(self, ctx:xDroneParser.EqualityContext):
        pass



del xDroneParser


from __future__ import annotations

import copy
import itertools
import math
import os
from itertools import combinations
from typing import List, Dict, Set, Tuple

import matplotlib.pyplot as plt
from scipy.stats import ncx2
from tabulate import tabulate

from xdrone.shared.collision_config import CollisionConfig
from xdrone.shared.command import AbstractDroneCommand, SingleDroneCommand, ParallelDroneCommands, Command
from xdrone.shared.drone_config import DroneConfig
from xdrone.shared.safety_check_error import SafetyCheckError
from xdrone.shared.state import State
from xdrone.state_updaters.state_updater import StateUpdater


class StateVariance:
    def __init__(self, state: State, variance: float = 0.0):
        self._state = state
        self._variance = variance

    @property
    def variance(self) -> float:
        return copy.deepcopy(self._variance)

    @property
    def state(self) -> State:
        return copy.deepcopy(self._state)


class CollisionChecker:
    def __init__(self, drone_config_map: Dict[str, DroneConfig], collision_config: CollisionConfig):
        self.drone_config_map = drone_config_map
        self.collision_config = collision_config

    def check(self, drone_commands: List[AbstractDroneCommand],
              state_updater_map: Dict[str, StateUpdater],
              save_report: bool = False):
        if len(self.drone_config_map) == 1:
            # no need to check if there is only 1 drone
            return
        drone_trajectory_map = {name: [StateVariance(state_updater.get_init_state())] for name, state_updater in
                                state_updater_map.items()}
        drones_involved = set(state_updater_map.keys())
        try:
            self._update_states_for_abstract_drone_command(drone_commands, state_updater_map, drone_trajectory_map,
                                                           drones_involved)
        except Exception as e:
            raise SafetyCheckError("Error occurred during collision check, "
                                   "please retry with a better collision_config. Error: " + str(e))

        collisions, time_slice_info = self._get_possible_collisions_and_time_slice_info(drone_trajectory_map)
        if save_report:
            CollisionReportSaver.save_check_report(time_slice_info)
        self._check_collisions(collisions)

    def _get_possible_collisions_and_time_slice_info(self, drone_trajectory_map: Dict[str, List[StateVariance]]) \
            -> (List[Tuple], List[Tuple]):
        time = 0
        time_slice_info = []
        collisions = []
        drone_trajectories = dict(drone_trajectory_map)
        while any(drone_trajectories.values()):
            time += self.collision_config.time_interval_seconds
            state_group = []
            for name, drone_trajectory in drone_trajectories.items():
                possible_states = []
                while drone_trajectory and drone_trajectory[0].state.time_used_seconds < time:
                    state_variance = drone_trajectory.pop(0)
                    possible_states.append((name, state_variance))
                if possible_states:
                    state_group.append(possible_states)
            # Commented out: check all possible states combinations if one drone has multiple states in time interval
            # possible_combinations = list(itertools.product(*state_group))
            # Only check combinations of last states if one drone has multiple states in time interval
            state_group = [possible_states[-1] for possible_states in state_group]
            possible_combinations = list(itertools.combinations(state_group, r=2))
            for possible_combination in possible_combinations:
                for (name1, state_variance1), (name2, state_variance2) in combinations(possible_combination, r=2):
                    state1, variance1 = state_variance1.state, state_variance1.variance
                    state2, variance2 = state_variance2.state, state_variance2.variance
                    variance = variance1 + variance2

                    x = state1.x_meters - state2.x_meters
                    y = state1.y_meters - state2.y_meters
                    z = state1.z_meters - state2.z_meters
                    mean_distance = math.sqrt(x ** 2 + y ** 2 + z ** 2)
                    collision_meters = self.collision_config.collision_meters

                    if math.isclose(variance, 0):
                        # when variance is 0, perform as determined, confidence is either 1 or 0
                        if collision_meters == 0:
                            confidence = 0.0
                        elif mean_distance > collision_meters + 1e-5:
                            confidence = 0.0
                        else:
                            confidence = 1.0
                    else:
                        # use non central chi-squared distribution to calculated confidence
                        nc = x ** 2 / variance + y ** 2 / variance + z ** 2 / variance
                        confidence = ncx2.cdf(collision_meters ** 2 / variance, df=3, nc=nc)

                    if confidence >= self.collision_config.confidence_threshold - 1e-8:
                        collisions.append((name1, name2, state1, state2, mean_distance, confidence))

                    time_slice_info.append((name1, name2, time, mean_distance, confidence))
        return collisions, time_slice_info

    def _check_collisions(self, collisions: List[Tuple[str, str, State, State, float, float]]):
        if collisions:
            error_msg = "Collisions might happen!\n"
            for name1, name2, state1, state2, mean_distance, confidence in collisions:
                time = round((state1.time_used_seconds + state2.time_used_seconds) / 2, 2)
                x = round((state1.x_meters + state2.x_meters) / 2, 2)
                y = round((state1.y_meters + state2.y_meters) / 2, 2)
                z = round((state1.z_meters + state2.z_meters) / 2, 2)
                mean_distance = round(mean_distance, 5)
                error_msg += ("Collision might happen between {} and {}, at time {}s, ".format(name1, name2, time) +
                              "near position (x={}m, y={}m, z={}m), distance={}m, confidence={:.3f}%\n"
                              .format(x, y, z, mean_distance, confidence * 100))
            raise SafetyCheckError(error_msg)

    def _update_states_for_abstract_drone_command(self, drone_commands: List[AbstractDroneCommand],
                                                  state_updaters: Dict[str, StateUpdater],
                                                  drone_trajectory_map: Dict[str, List[StateVariance]],
                                                  drones_involved: Set[str]) -> float:
        total_time_used = 0
        for drone_command in drone_commands:
            if isinstance(drone_command, SingleDroneCommand):
                total_time_used += self._update_states_for_single_drone_command(drone_command,
                                                                                state_updaters,
                                                                                drone_trajectory_map,
                                                                                drones_involved)
            elif isinstance(drone_command, ParallelDroneCommands):
                total_time_used += self._update_states_for_parallel_drone_commands(drone_command,
                                                                                   state_updaters,
                                                                                   drone_trajectory_map,
                                                                                   drones_involved)
        return total_time_used

    def _update_states_for_single_drone_command(self, single_drone_command: SingleDroneCommand,
                                                state_updaters: Dict[str, StateUpdater],
                                                drone_trajectory_map: Dict[str, List[StateVariance]],
                                                drones_involved: Set[str]) -> float:
        time_used = self._update_states(single_drone_command, state_updaters, drone_trajectory_map, drones_involved)
        return time_used

    def _get_last_state_and_variance(self, drone_trajectory_map: Dict[str, List[StateVariance]],
                                     drone_name: str) -> (State, float):
        old_state_variance = drone_trajectory_map[drone_name][-1]
        old_state = old_state_variance.state
        old_variance = old_state_variance.variance
        return old_state, old_variance

    def _update_states(self, single_drone_command: SingleDroneCommand,
                       state_updaters: Dict[str, StateUpdater],
                       drone_trajectory_map: Dict[str, List[StateVariance]],
                       drones_involved: Set[str]) -> float:
        time_interval = self.collision_config.time_interval_seconds
        drone_name = single_drone_command.drone_name
        command = single_drone_command.command
        if command.opcode == "takeoff":
            takeoff_distance = self.drone_config_map[drone_name].takeoff_height_meters
            new_drone_command = SingleDroneCommand(drone_name, Command.up(takeoff_distance))
            return self._update_states(new_drone_command, state_updaters, drone_trajectory_map, drones_involved)
        if command.opcode == "land":
            old_state, old_variance = self._get_last_state_and_variance(drone_trajectory_map, drone_name)
            land_distance = old_state.z_meters
            new_drone_command = SingleDroneCommand(drone_name, Command.down(land_distance))
            return self._update_states(new_drone_command, state_updaters, drone_trajectory_map, drones_involved)
        if command.opcode == "wait":
            seconds, = command.operands
            old_state, old_variance = self._get_last_state_and_variance(drone_trajectory_map, drone_name)
            if seconds <= time_interval:
                new_state = state_updaters[drone_name].update(Command.wait(seconds), old_state)
                new_state_variance = StateVariance(new_state, old_variance)
                drone_trajectory_map[drone_name].append(new_state_variance)
                for name in drones_involved.difference({drone_name}):
                    state, variance = self._get_last_state_and_variance(drone_trajectory_map, name)
                    state = state.copy_and_set_time_used_seconds(state.time_used_seconds + seconds)
                    drone_trajectory_map[name].append(StateVariance(state, variance))
                return seconds
            else:
                new_state = state_updaters[drone_name].update(Command.wait(time_interval), old_state)
                new_state_variance = StateVariance(new_state, old_variance)
                drone_trajectory_map[drone_name].append(new_state_variance)
                for name in drones_involved.difference({drone_name}):
                    state, variance = self._get_last_state_and_variance(drone_trajectory_map, name)
                    state = state.copy_and_set_time_used_seconds(state.time_used_seconds + time_interval)
                    drone_trajectory_map[name].append(StateVariance(state, variance))
                new_drone_command = SingleDroneCommand(drone_name, Command.wait(seconds - time_interval))
                return time_interval + self._update_states(new_drone_command, state_updaters, drone_trajectory_map,
                                                           drones_involved)
        if command.opcode in ["rotate_left", "rotate_right"]:
            degrees, = command.operands
            old_state, old_variance = self._get_last_state_and_variance(drone_trajectory_map, drone_name)
            rotate_speed = self.drone_config_map[drone_name].rotate_speed_dps
            seconds = degrees / rotate_speed
            if seconds <= time_interval:
                new_state = state_updaters[drone_name].update(Command(command.opcode, [rotate_speed * seconds]),
                                                              old_state)
                new_variance = old_variance + self.drone_config_map[drone_name].var_per_degree * seconds * rotate_speed
                new_state_variance = StateVariance(new_state, new_variance)
                drone_trajectory_map[drone_name].append(new_state_variance)
                for name in drones_involved.difference({drone_name}):
                    state, variance = self._get_last_state_and_variance(drone_trajectory_map, name)
                    state = state.copy_and_set_time_used_seconds(state.time_used_seconds + seconds)
                    drone_trajectory_map[name].append(StateVariance(state, variance))
                return seconds
            else:
                new_state = state_updaters[drone_name].update(Command(command.opcode, [rotate_speed * time_interval]),
                                                              old_state)
                new_variance = old_variance + \
                               self.drone_config_map[drone_name].var_per_degree * time_interval * rotate_speed
                new_state_variance = StateVariance(new_state, new_variance)
                drone_trajectory_map[drone_name].append(new_state_variance)
                for name in drones_involved.difference({drone_name}):
                    state, variance = self._get_last_state_and_variance(drone_trajectory_map, name)
                    state = state.copy_and_set_time_used_seconds(state.time_used_seconds + time_interval)
                    drone_trajectory_map[name].append(StateVariance(state, variance))
                new_drone_command = SingleDroneCommand(drone_name,
                                                       Command(command.opcode,
                                                               [degrees - rotate_speed * time_interval]))
                return time_interval + self._update_states(new_drone_command, state_updaters, drone_trajectory_map,
                                                           drones_involved)
        if command.opcode in ["up", "down", "left", "right", "forward", "backward"]:
            meters, = command.operands
            old_state_variance = drone_trajectory_map[drone_name][-1]
            old_state = old_state_variance.state
            old_variance = old_state_variance.variance
            speed = self.drone_config_map[drone_name].speed_mps
            seconds = meters / speed
            if seconds <= time_interval:
                new_state = state_updaters[drone_name].update(Command(command.opcode, [speed * seconds]),
                                                              old_state)
                new_variance = old_variance + self.drone_config_map[drone_name].var_per_meter * seconds * speed
                new_state_variance = StateVariance(new_state, new_variance)
                drone_trajectory_map[drone_name].append(new_state_variance)
                for name in drones_involved.difference({drone_name}):
                    state_variance = drone_trajectory_map[name][-1]
                    state = state_variance.state
                    variance = state_variance.variance
                    state = state.copy_and_set_time_used_seconds(state.time_used_seconds + seconds)
                    drone_trajectory_map[name].append(StateVariance(state, variance))
                return seconds
            else:
                new_state = state_updaters[drone_name].update(Command(command.opcode, [speed * time_interval]),
                                                              old_state)
                new_variance = old_variance + self.drone_config_map[drone_name].var_per_meter * time_interval * speed
                new_state_variance = StateVariance(new_state, new_variance)
                drone_trajectory_map[drone_name].append(new_state_variance)
                for name in drones_involved.difference({drone_name}):
                    state_variance = drone_trajectory_map[name][-1]
                    state = state_variance.state
                    variance = state_variance.variance
                    state = state.copy_and_set_time_used_seconds(state.time_used_seconds + time_interval)
                    drone_trajectory_map[name].append(StateVariance(state, variance))
                new_drone_command = SingleDroneCommand(drone_name,
                                                       Command(command.opcode, [meters - speed * time_interval]))
                return time_interval + self._update_states(new_drone_command, state_updaters, drone_trajectory_map,
                                                           drones_involved)

    def _update_states_for_parallel_drone_commands(self, parallel_drone_commands: ParallelDroneCommands,
                                                   state_updaters: Dict[str, StateUpdater],
                                                   drone_trajectory_map: Dict[str, List[StateVariance]],
                                                   drones_involved: Set[str]) -> float:
        assert len(parallel_drone_commands.branches) > 0
        time_used_in_branches = self._update_states_and_check_for_each_branch(parallel_drone_commands,
                                                                              state_updaters,
                                                                              drone_trajectory_map)
        longest_time_used = self._update_states_to_wait_for_slowest_branch(parallel_drone_commands,
                                                                           state_updaters,
                                                                           drone_trajectory_map,
                                                                           time_used_in_branches,
                                                                           drones_involved)
        return longest_time_used

    def _update_states_and_check_for_each_branch(self, parallel_drone_commands: ParallelDroneCommands,
                                                 state_updaters: Dict[str, StateUpdater],
                                                 drone_trajectory_map: Dict[str, List[StateVariance]]) -> List[float]:
        time_used_in_branches = []
        for i, branch in enumerate(parallel_drone_commands.branches):
            drones_involved = parallel_drone_commands.drones_involved_each_branch[i]
            time_used = self._update_states_for_abstract_drone_command(branch, state_updaters,
                                                                       drone_trajectory_map, drones_involved)
            time_used_in_branches.append(time_used)
        return time_used_in_branches

    def _update_states_to_wait_for_slowest_branch(self, parallel_drone_commands: ParallelDroneCommands,
                                                  state_updaters: Dict[str, StateUpdater],
                                                  drone_trajectory_map: Dict[str, List[StateVariance]],
                                                  time_used_in_branches: List[float],
                                                  drones_involved: Set[str]) -> float:
        longest_time_used = max(time_used_in_branches)
        # for each branch, let drones involved in the branch wait until longest_time_used
        for i, time_used in enumerate(time_used_in_branches):
            for name in parallel_drone_commands.drones_involved_each_branch[i]:
                wait_command = Command.wait(longest_time_used - time_used)
                self._update_states_for_single_drone_command(SingleDroneCommand(name, wait_command),
                                                             state_updaters,
                                                             drone_trajectory_map,
                                                             drones_involved={name})
        # let drones not involved in any branch wait for longest_time_used
        for name in drones_involved.difference(parallel_drone_commands.get_drones_involved()):
            wait_command = Command.wait(longest_time_used)
            self._update_states_for_single_drone_command(SingleDroneCommand(name, wait_command),
                                                         state_updaters,
                                                         drone_trajectory_map,
                                                         drones_involved={name})
        return longest_time_used


class CollisionReportSaver:
    @staticmethod
    def save_check_report(time_slice_info: List[Tuple[str, str, float, float, float]]):
        report_dir = "reports"
        if not os.path.isdir(report_dir):
            os.mkdir(report_dir)

        file_name = os.path.join(report_dir, "collision check log (all drones).txt")

        with open(file_name, "w") as file:
            sorted_time_slice_info = sorted(time_slice_info, key=lambda elem: elem[2])
            file.write(tabulate(CollisionReportSaver._format_rows(sorted_time_slice_info),
                                headers=["Drone 1", "Drone 2", "Time", "Distance", "Confidence"],
                                colalign=("default", "default", "default", "default", "right")))

        drone_pairs = sorted(set([(name1, name2) for (name1, name2, _, _, _) in time_slice_info]))

        for drone_pair in drone_pairs:
            filtered_time_slice_info = filter(lambda elem: drone_pair == (elem[0], elem[1]), time_slice_info)
            filtered_time_slice_info = sorted(filtered_time_slice_info, key=lambda elem: elem[2])

            file_name = os.path.join(report_dir, "collision check log ({}-{}).txt".format(drone_pair[0], drone_pair[1]))
            with open(file_name, "w") as file:
                file.write(tabulate(CollisionReportSaver._format_rows(filtered_time_slice_info),
                                    headers=["Drone 1", "Drone 2", "Time", "Distance", "Confidence"],
                                    colalign=("default", "default", "default", "default", "right")))

            times = [time for (_, _, time, _, _) in filtered_time_slice_info]
            probabilities = [prob for (_, _, _, _, prob) in filtered_time_slice_info]
            fig, ax = plt.subplots()
            ax.plot(times, probabilities)
            ax.set_xlabel("Time")
            ax.set_ylabel("Probability")
            ax.set_title("Probability of Collision between {} and {}".format(drone_pair[0], drone_pair[1]))
            ax.set_ylim([-0.05, 1.05])
            fig_name = os.path.join(report_dir, "collision probability ({}-{}).png".format(drone_pair[0], drone_pair[1]))
            fig.savefig(fig_name)

    @staticmethod
    def _format_rows(table: List[Tuple[str, str, float, float, float]]) -> List[Tuple[str, str, float, float, str]]:
        return [(name1, name2, round(time, 3), round(distance, 5), "{:3.3f}%".format(confidence * 100))
                for (name1, name2, time, distance, confidence) in table]



import unittest

from xdrone.shared.collision_config import CollisionConfig, DefaultCollisionConfig


class CollisionConfigTest(unittest.TestCase):
    def test_init_with_wrong_parameter_should_give_error(self):
        for invalid_value in [-10, -1]:
            with self.assertRaises(ValueError) as context:
                CollisionConfig(collision_meters=invalid_value)
            self.assertTrue("collision_meters should >= 0" in str(context.exception))
        for invalid_value in [-10, -1, 0]:
            with self.assertRaises(ValueError) as context:
                CollisionConfig(time_interval_seconds=invalid_value)
            self.assertTrue("time_interval_seconds should > 0" in str(context.exception))
        for invalid_value in [-10, -1, -0.1, 1.1, 10]:
            with self.assertRaises(ValueError) as context:
                CollisionConfig(confidence_threshold=invalid_value)
            self.assertTrue("confidence_threshold should between 0 and 1" in str(context.exception))

    def test_property(self):
        drone_config = CollisionConfig(collision_meters=0.3, time_interval_seconds=0.5, confidence_threshold=0.95)
        self.assertEqual(0.3, drone_config.collision_meters)
        self.assertEqual(0.5, drone_config.time_interval_seconds)
        self.assertEqual(0.95, drone_config.confidence_threshold)

    def test_str(self):
        drone_config = CollisionConfig(collision_meters=0.3, time_interval_seconds=0.5, confidence_threshold=0.95)
        self.assertEqual("CollisionConfig: { collision_meters: 0.3, time_interval_seconds: 0.5, " +
                         "confidence_threshold: 0.95 }",
                         str(drone_config))

    def test_eq(self):
        self.assertEqual(CollisionConfig(),
                         CollisionConfig())
        self.assertEqual(CollisionConfig(collision_meters=0.3, time_interval_seconds=0.5, confidence_threshold=0.95),
                         CollisionConfig(collision_meters=0.3, time_interval_seconds=0.5, confidence_threshold=0.95))
        self.assertNotEqual(CollisionConfig(collision_meters=0.1, time_interval_seconds=0.5, confidence_threshold=0.95),
                            CollisionConfig(collision_meters=0.3, time_interval_seconds=0.5, confidence_threshold=0.95))
        self.assertNotEqual(CollisionConfig(collision_meters=0.3, time_interval_seconds=0.1, confidence_threshold=0.95),
                            CollisionConfig(collision_meters=0.3, time_interval_seconds=0.5, confidence_threshold=0.95))
        self.assertNotEqual(CollisionConfig(collision_meters=0.3, time_interval_seconds=0.5, confidence_threshold=0.95),
                         CollisionConfig(collision_meters=0.3, time_interval_seconds=0.5, confidence_threshold=0.9))
        self.assertNotEqual(None,
                            CollisionConfig(collision_meters=0.3, time_interval_seconds=0.5, confidence_threshold=0.95))


class DefaultCollisionConfigTest(unittest.TestCase):
    def test_default_values(self):
        collision_config = DefaultCollisionConfig()
        self.assertEqual(0, collision_config.collision_meters)
        self.assertEqual(0.1, collision_config.time_interval_seconds)
        self.assertEqual(1.0, collision_config.confidence_threshold)



import unittest

from xdrone.compiler.compiler_utils.functions import FunctionIdentifier, Parameter, Function, FunctionTable
from xdrone.compiler.compiler_utils.type import Type


class TestFunctionIdentifier(unittest.TestCase):
    def test_property(self):
        self.assertEqual("a", FunctionIdentifier("a").ident)

    def test_str(self):
        self.assertEqual("FunctionIdentifier: { ident: a }", str(FunctionIdentifier("a")))

    def test_eq(self):
        self.assertEqual(FunctionIdentifier("a"), FunctionIdentifier("a"))
        self.assertNotEqual(FunctionIdentifier("a"), FunctionIdentifier("abc"))
        self.assertNotEqual(FunctionIdentifier("a"), None)


class TestParameter(unittest.TestCase):
    def test_property(self):
        self.assertEqual("a", Parameter("a", Type.int()).ident)
        self.assertEqual(Type.int(), Parameter("a", Type.int()).type)

    def test_str(self):
        self.assertEqual("Parameter: { ident: a, type: int }", str(Parameter("a", Type.int())))

    def test_eq(self):
        self.assertEqual(Parameter("a", Type.int()), Parameter("a", Type.int()))
        self.assertNotEqual(Parameter("a", Type.int()), Parameter("abc", Type.int()))
        self.assertNotEqual(Parameter("a", Type.int()), Parameter("a", Type.list_of(Type.int())))
        self.assertNotEqual(Parameter("a", Type.int()), None)


class TestFunction(unittest.TestCase):
    def test_property(self):
        self.assertEqual("a", Function("a", [], Type.int(), []).ident)
        self.assertEqual([Parameter("a", Type.int())],
                         Function("a", [Parameter("a", Type.int())], Type.int(), []).param_list)
        self.assertEqual(Type.int(), Function("a", [], Type.int(), []).return_type)

    def test_get_commands(self):
        self.assertEqual([], Function("a", [Parameter("a", Type.int())], Type.int(), []).get_commands())

    def test_str(self):
        self.assertEqual("Function: { ident: a, param_list: [], return_type: int }",
                         str(Function("a", [], Type.int(), [])))
        self.assertEqual("Function: { ident: a, param_list: [Parameter: { ident: a, type: int }, "
                         "Parameter: { ident: b, type: string }], "
                         "return_type: None }",
                         str(Function("a", [Parameter("a", Type.int()), Parameter("b", Type.string())], None, [])))

    def test_eq(self):
        self.assertEqual(Function("a", [Parameter("a", Type.int()), Parameter("b", Type.string())], None, []),
                         Function("a", [Parameter("a", Type.int()), Parameter("b", Type.string())], None, []))
        self.assertNotEqual(Function("a", [Parameter("a", Type.int()), Parameter("b", Type.string())], None, []),
                            Function("b", [Parameter("a", Type.int()), Parameter("b", Type.string())], None, []))
        self.assertNotEqual(Function("a", [Parameter("a", Type.int()), Parameter("b", Type.string())], None, []),
                            Function("a", [Parameter("a", Type.int())], None, []))
        self.assertNotEqual(Function("a", [Parameter("a", Type.int()), Parameter("b", Type.string())], None, []),
                            Function("a", [Parameter("a", Type.int()), Parameter("b", Type.string())], Type.int(), []))
        # Note that `commands` will not be compared
        self.assertEqual(Function("a", [Parameter("a", Type.int())], None, []),
                         Function("a", [Parameter("a", Type.int())], None, ['dummy command']))
        self.assertNotEqual(Function("a", [Parameter("a", Type.int())], None, []),
                            None)


class TestFunctionTable(unittest.TestCase):
    def test_str(self):
        ft = FunctionTable()
        ft.store("a", Function("a", [], None, []))
        expected = "FunctionTable: {\n" \
                   + "  a -> Function: { ident: a, param_list: [], return_type: None }\n" \
                   + "}"
        self.assertEqual(expected, str(ft))

    def test_eq(self):
        ft1 = FunctionTable()
        ft2 = FunctionTable()
        self.assertEqual(ft1, ft2)
        ft1.store("a", Function("a", [Parameter("a", Type.int()), Parameter("b", Type.string())], None, []))
        self.assertNotEqual(ft1, ft2)
        # Note that `commands` in Function will not be compared
        ft2.store("a", Function("a", [Parameter("a", Type.int()), Parameter("b", Type.string())], None, ["dummy"]))
        self.assertEqual(ft1, ft2)

        self.assertNotEqual(FunctionTable(), None)

    def test_contains(self):
        ft = FunctionTable()
        self.assertFalse("a" in ft)
        ft.store("a", Function("a", [Parameter("a", Type.int()), Parameter("b", Type.string())], None, []))
        self.assertTrue("a" in ft)

    def test_store(self):
        ft = FunctionTable()
        ft.store("a", Function("a", [Parameter("a", Type.int()), Parameter("b", Type.string())], None, []))
        self.assertTrue("a" in ft)
        self.assertEqual(Function("a", [Parameter("a", Type.int()), Parameter("b", Type.string())], None, []),
                         ft.get_function("a"))
        self.assertRaises(AssertionError, ft.store, "a",
                          Function("a", [Parameter("a", Type.int()), Parameter("b", Type.string())], None, []))

    def test_get_function(self):
        ft = FunctionTable()
        self.assertRaises(AssertionError, ft.get_function, "a", )
        ft.store("a", Function("a", [Parameter("a", Type.int()), Parameter("b", Type.string())], None, []))
        self.assertTrue("a" in ft)
        self.assertEqual(Function("a", [Parameter("a", Type.int()), Parameter("b", Type.string())], None, []),
                         ft.get_function("a"))



import logging
import unittest
from unittest.mock import Mock, call

from xdrone.command_converters.dji_tello_drone_executor import DJITelloExecutor
from xdrone.shared.command import Command


class TestDJITelloExecutor(unittest.TestCase):

    def test_dji_tello_executor_should_send_correct_message(self) -> None:
        mocked_socket = Mock()
        mocked_socket.recvfrom = Mock(return_value=(b'ok', ('192.168.10.1', 8889)))
        mocked_socket.sendto = Mock()
        commands = [Command.takeoff(),
                    Command.up(1),
                    Command.down(1),
                    Command.left(1),
                    Command.right(1),
                    Command.forward(1),
                    Command.backward(1),
                    Command.rotate_left(90),
                    Command.rotate_right(90),
                    Command.wait(0),
                    Command.land()]
        executor = DJITelloExecutor()
        executor.sock.close()
        executor.sock = mocked_socket
        with self.assertLogs(logging.getLogger()) as log:
            executor.execute_commands(commands)
        calls = [call(b"command", ('192.168.10.1', 8889)),
                 call(b"takeoff", ('192.168.10.1', 8889)),
                 call(b"up 100", ('192.168.10.1', 8889)),
                 call(b"down 100", ('192.168.10.1', 8889)),
                 call(b"left 100", ('192.168.10.1', 8889)),
                 call(b"right 100", ('192.168.10.1', 8889)),
                 call(b"forward 100", ('192.168.10.1', 8889)),
                 call(b"back 100", ('192.168.10.1', 8889)),
                 call(b"ccw 90", ('192.168.10.1', 8889)),
                 call(b"cw 90", ('192.168.10.1', 8889)),
                 call(b"land", ('192.168.10.1', 8889))]
        mocked_socket.sendto.assert_has_calls(calls)
        expected_log = ["INFO:root:sent message: b'command'",
                        "INFO:root:received response: ok",
                        "INFO:root:sent message: b'takeoff'",
                        "INFO:root:received response: ok",
                        "INFO:root:sent message: b'up 100'",
                        "INFO:root:received response: ok",
                        "INFO:root:sent message: b'down 100'",
                        "INFO:root:received response: ok",
                        "INFO:root:sent message: b'left 100'",
                        "INFO:root:received response: ok",
                        "INFO:root:sent message: b'right 100'",
                        "INFO:root:received response: ok",
                        "INFO:root:sent message: b'forward 100'",
                        "INFO:root:received response: ok",
                        "INFO:root:sent message: b'back 100'",
                        "INFO:root:received response: ok",
                        "INFO:root:sent message: b'ccw 90'",
                        "INFO:root:received response: ok",
                        "INFO:root:sent message: b'cw 90'",
                        "INFO:root:received response: ok",
                        "INFO:root:sent message: b'land'",
                        "INFO:root:received response: ok"]
        self.assertEqual(expected_log, log.output)

    def test_dji_tello_executor_if_receive_error_should_print(self) -> None:
        mocked_socket = Mock()
        mocked_socket.recvfrom = Mock(side_effect=Exception('timed out'))
        mocked_socket.sendto = Mock()
        commands = [Command.takeoff(),
                    Command.land()]
        executor = DJITelloExecutor()
        executor.sock.close()
        executor.sock = mocked_socket
        with self.assertLogs(logging.getLogger()) as log:
            executor.execute_commands(commands)
        calls = [call(b"command", ('192.168.10.1', 8889)),
                 call(b"takeoff", ('192.168.10.1', 8889)),
                 call(b"land", ('192.168.10.1', 8889))]
        mocked_socket.sendto.assert_has_calls(calls)
        expected_log = ["INFO:root:sent message: b'command'",
                        "ERROR:root:Error met when receiving response: timed out",
                        "INFO:root:sent message: b'takeoff'",
                        "ERROR:root:Error met when receiving response: timed out",
                        "INFO:root:sent message: b'land'",
                        "ERROR:root:Error met when receiving response: timed out"]
        self.assertEqual(expected_log, log.output)

    def test_dji_tello_executor_if_interrupt_error_should_stop_and_emergency(self) -> None:
        mocked_socket = Mock()
        mocked_socket.recvfrom = Mock(return_value=(b'ok', ('192.168.10.1', 8889)))
        mocked_socket.sendto = Mock()

        def mocked_sendto(msg, _):
            if msg == b"takeoff":
                raise KeyboardInterrupt

        mocked_socket.sendto.side_effect = mocked_sendto
        commands = [Command.takeoff(),
                    Command.land()]
        executor = DJITelloExecutor()
        executor.sock.close()
        executor.sock = mocked_socket
        with self.assertLogs(logging.getLogger()) as log:
            executor.execute_commands(commands)
        calls = [call(b"command", ('192.168.10.1', 8889)),
                 call(b"takeoff", ('192.168.10.1', 8889)),
                 call(b"emergency", ('192.168.10.1', 8889))]
        mocked_socket.sendto.assert_has_calls(calls)
        expected_log = ["INFO:root:sent message: b'command'",
                        "INFO:root:received response: ok",
                        "INFO:root:KeyboardInterrupt received. Forced stop."]
        self.assertEqual(expected_log, log.output)

    def test_dji_tello_executor_if_interrupted_before_recv_thread_start_should_not_give_error(self) -> None:
        commands = [Command.takeoff(),
                    Command.land()]
        executor = DJITelloExecutor()
        executor.recv_thread.start = Mock(side_effect=KeyboardInterrupt)
        executor.execute_commands(commands)



import copy

from xdrone.shared.drone_config import DroneConfig, DefaultDroneConfig


class Drone:
    def __init__(self, name: str, config: DroneConfig):
        self._name = name
        self._config = config

    @property
    def name(self) -> str:
        return copy.deepcopy(self._name)

    @property
    def config(self) -> DroneConfig:
        return copy.deepcopy(self._config)

    def __str__(self):
        return "Drone: {{ name: {}, config: {} }}".format(self._name, self._config)

    def __eq__(self, other):
        if isinstance(other, Drone):
            return other._name == self._name and other._config == self._config
        return False


class NullDrone(Drone):
    def __init__(self):
        super().__init__("null", DefaultDroneConfig())

    def __str__(self):
        return "NullDrone: { }"

    def __eq__(self, other):
        return isinstance(other, NullDrone)



import copy
from typing import Union, Set

from xdrone.compiler.compiler_utils.drones import Drone
from xdrone.compiler.compiler_utils.expressions import Expression


class SymbolTable:
    def __init__(self):
        self._symbol_table = {}

    def __str__(self):
        result = "SymbolTable: {\n"
        for ident, expression in self._symbol_table.items():
            result += "  " + ident + " -> " + str(expression) + "\n"
        result += "}"
        return result

    def __eq__(self, other):
        if isinstance(other, SymbolTable):
            return other._symbol_table == self._symbol_table
        return False

    def __contains__(self, ident: str) -> bool:
        return ident in self._symbol_table

    def store(self, ident: str, expression: Expression) -> None:
        assert not self.__contains__(ident), "Expression already in symbol table, please use update() to update it"
        self._symbol_table[ident] = expression

    def update(self, ident: str, value: Union[int, float, str, bool, Drone, list]) -> None:
        assert self.__contains__(ident), "Expression not in symbol table, please use store() to store it first"
        old_expression = self._symbol_table[ident]
        self._symbol_table[ident] = Expression(old_expression.type, value, ident=ident)

    def delete(self, ident: str) -> None:
        assert self.__contains__(ident), "Expression not in symbol table"
        del self._symbol_table[ident]

    def get_expression(self, ident: str) -> Expression:
        assert self.__contains__(ident), "Expression not in symbol table"
        return copy.deepcopy(self._symbol_table[ident])

    def idents(self) -> Set[str]:
        return set(self._symbol_table.keys())

    def keep_idents(self, idents: Set[str]) -> None:
        self._symbol_table = {ident: self._symbol_table[ident] for ident in idents}



from __future__ import annotations

import copy
from typing import Optional, Union

from xdrone.compiler.compiler_utils.drones import Drone
from xdrone.compiler.compiler_utils.type import ListType, Type


class AbstractExpression:
    def to_expression(self) -> Optional[Expression]:
        raise NotImplementedError("to_expression not supported in AbstractExpression")


class Expression(AbstractExpression):
    def __init__(self, type: Type, value: Union[int, float, str, bool, Drone, list], ident: Optional[str] = None):
        self._type = type
        self._value = value
        self._ident = ident

    @property
    def type(self) -> Type:
        return copy.deepcopy(self._type)

    @property
    def value(self) -> Union[int, float, str, bool, Drone, list]:
        return copy.deepcopy(self._value)

    @property
    def ident(self) -> Optional[str]:
        return copy.deepcopy(self._ident)

    def to_expression(self) -> Optional[Expression]:
        return Expression(self._type, self._value, self._ident)

    def __str__(self):
        return "Expression: {{ type: {}, value: {}, ident: {} }}".format(self._type, self._value, self._ident)

    def __eq__(self, other):
        if isinstance(other, Expression):
            return other._type == self._type and other._value == self._value and other.ident == self._ident
        return False


class Identifier(AbstractExpression):
    def __init__(self, ident: str, expression: Optional[Expression]):
        self._ident = ident
        self._expression = expression

    @property
    def ident(self) -> str:
        return copy.deepcopy(self._ident)

    def to_expression(self) -> Optional[Expression]:
        return copy.deepcopy(self._expression)

    def __str__(self):
        return "Identifier: {{ ident: {}, expression: {} }}".format(self._ident, self._expression)

    def __eq__(self, other):
        if isinstance(other, Identifier):
            return other._ident == self._ident and other._expression == self._expression
        return False


class ListElem(AbstractExpression):
    def __init__(self, ident: str, container: Expression, index: int):
        assert isinstance(container.type, ListType)
        self._ident = ident
        self._container = container
        self._index = index
        self._expression = Expression(container.type.elem_type, container.value[index],
                                      ident="{}[{}]".format(ident, index))

    @property
    def ident(self) -> str:
        return copy.deepcopy(self._ident)

    @property
    def container(self) -> Expression:
        return copy.deepcopy(self._container)

    @property
    def index(self) -> int:
        return copy.deepcopy(self._index)

    def to_expression(self) -> Optional[Expression]:
        return copy.deepcopy(self._expression)

    def __str__(self):
        return "ListElem: {{ ident: {}, container: {}, index: {}, expression: {} }}".format(
            self._ident, self._container, self.index, self._expression)

    def __eq__(self, other):
        if isinstance(other, ListElem):
            return other._ident == self._ident and other._container == self._container \
                   and other._index == self._index and other._expression == self._expression
        return False


class VectorElem(AbstractExpression):
    def __init__(self, ident: str, container: Expression, index: int):
        assert container.type == Type.vector()
        self._ident = ident
        self._container = container
        self._index = index
        self._expression = Expression(Type.decimal(), container.value[index])

    @property
    def ident(self) -> str:
        return copy.deepcopy(self._ident)

    @property
    def container(self) -> Expression:
        return copy.deepcopy(self._container)

    @property
    def index(self) -> int:
        return copy.deepcopy(self._index)

    def to_expression(self) -> Optional[Expression]:
        return copy.deepcopy(self._expression)

    def __str__(self):
        return "ListElem: {{ ident: {}, container: {}, index: {}, expression: {} }}".format(
            self._ident, self._container, self.index, self._expression)

    def __eq__(self, other):
        if isinstance(other, VectorElem):
            return other._ident == self._ident and other._container == self._container \
                   and other._index == self._index and other._expression == self._expression
        return False



# Generated from xDroneParser.g4 by ANTLR 4.9.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .xDroneParser import xDroneParser
else:
    from xDroneParser import xDroneParser

# This class defines a complete generic visitor for a parse tree produced by xDroneParser.

class xDroneParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by xDroneParser#prog.
    def visitProg(self, ctx:xDroneParser.ProgContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#commands.
    def visitCommands(self, ctx:xDroneParser.CommandsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#takeoff.
    def visitTakeoff(self, ctx:xDroneParser.TakeoffContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#land.
    def visitLand(self, ctx:xDroneParser.LandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#up.
    def visitUp(self, ctx:xDroneParser.UpContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#down.
    def visitDown(self, ctx:xDroneParser.DownContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#left.
    def visitLeft(self, ctx:xDroneParser.LeftContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#right.
    def visitRight(self, ctx:xDroneParser.RightContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#forward.
    def visitForward(self, ctx:xDroneParser.ForwardContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#backward.
    def visitBackward(self, ctx:xDroneParser.BackwardContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#rotateLeft.
    def visitRotateLeft(self, ctx:xDroneParser.RotateLeftContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#rotateRight.
    def visitRotateRight(self, ctx:xDroneParser.RotateRightContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#wait.
    def visitWait(self, ctx:xDroneParser.WaitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#declare.
    def visitDeclare(self, ctx:xDroneParser.DeclareContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#declareAssign.
    def visitDeclareAssign(self, ctx:xDroneParser.DeclareAssignContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#assignVectorElem.
    def visitAssignVectorElem(self, ctx:xDroneParser.AssignVectorElemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#assignListElem.
    def visitAssignListElem(self, ctx:xDroneParser.AssignListElemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#assignIdent.
    def visitAssignIdent(self, ctx:xDroneParser.AssignIdentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#del.
    def visitDel(self, ctx:xDroneParser.DelContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#insert.
    def visitInsert(self, ctx:xDroneParser.InsertContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#remove.
    def visitRemove(self, ctx:xDroneParser.RemoveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#procedureCall.
    def visitProcedureCall(self, ctx:xDroneParser.ProcedureCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#if.
    def visitIf(self, ctx:xDroneParser.IfContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#while.
    def visitWhile(self, ctx:xDroneParser.WhileContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#for.
    def visitFor(self, ctx:xDroneParser.ForContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#repeat.
    def visitRepeat(self, ctx:xDroneParser.RepeatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#return.
    def visitReturn(self, ctx:xDroneParser.ReturnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#parallel.
    def visitParallel(self, ctx:xDroneParser.ParallelContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#ident.
    def visitIdent(self, ctx:xDroneParser.IdentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#listElem.
    def visitListElem(self, ctx:xDroneParser.ListElemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#vectorX.
    def visitVectorX(self, ctx:xDroneParser.VectorXContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#vectorY.
    def visitVectorY(self, ctx:xDroneParser.VectorYContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#vectorZ.
    def visitVectorZ(self, ctx:xDroneParser.VectorZContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#funcIdent.
    def visitFuncIdent(self, ctx:xDroneParser.FuncIdentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#call.
    def visitCall(self, ctx:xDroneParser.CallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#argList.
    def visitArgList(self, ctx:xDroneParser.ArgListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#function.
    def visitFunction(self, ctx:xDroneParser.FunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#procedure.
    def visitProcedure(self, ctx:xDroneParser.ProcedureContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#paramList.
    def visitParamList(self, ctx:xDroneParser.ParamListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#intType.
    def visitIntType(self, ctx:xDroneParser.IntTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#decimalType.
    def visitDecimalType(self, ctx:xDroneParser.DecimalTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#stringType.
    def visitStringType(self, ctx:xDroneParser.StringTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#booleanType.
    def visitBooleanType(self, ctx:xDroneParser.BooleanTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#vectorType.
    def visitVectorType(self, ctx:xDroneParser.VectorTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#droneType.
    def visitDroneType(self, ctx:xDroneParser.DroneTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#listType.
    def visitListType(self, ctx:xDroneParser.ListTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#parentheses.
    def visitParentheses(self, ctx:xDroneParser.ParenthesesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#plusMinus.
    def visitPlusMinus(self, ctx:xDroneParser.PlusMinusContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#compare.
    def visitCompare(self, ctx:xDroneParser.CompareContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#intExpr.
    def visitIntExpr(self, ctx:xDroneParser.IntExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#or.
    def visitOr(self, ctx:xDroneParser.OrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#decimalExpr.
    def visitDecimalExpr(self, ctx:xDroneParser.DecimalExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#trueExpr.
    def visitTrueExpr(self, ctx:xDroneParser.TrueExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#listElemExpr.
    def visitListElemExpr(self, ctx:xDroneParser.ListElemExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#positNegate.
    def visitPositNegate(self, ctx:xDroneParser.PositNegateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#concat.
    def visitConcat(self, ctx:xDroneParser.ConcatContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#list.
    def visitList(self, ctx:xDroneParser.ListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#vectorYExpr.
    def visitVectorYExpr(self, ctx:xDroneParser.VectorYExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#stringExpr.
    def visitStringExpr(self, ctx:xDroneParser.StringExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#identExpr.
    def visitIdentExpr(self, ctx:xDroneParser.IdentExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#not.
    def visitNot(self, ctx:xDroneParser.NotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#vectorXExpr.
    def visitVectorXExpr(self, ctx:xDroneParser.VectorXExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#size.
    def visitSize(self, ctx:xDroneParser.SizeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#and.
    def visitAnd(self, ctx:xDroneParser.AndContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#functionCall.
    def visitFunctionCall(self, ctx:xDroneParser.FunctionCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#vectorZExpr.
    def visitVectorZExpr(self, ctx:xDroneParser.VectorZExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#vector.
    def visitVector(self, ctx:xDroneParser.VectorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#falseExpr.
    def visitFalseExpr(self, ctx:xDroneParser.FalseExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#multiDivide.
    def visitMultiDivide(self, ctx:xDroneParser.MultiDivideContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by xDroneParser#equality.
    def visitEquality(self, ctx:xDroneParser.EqualityContext):
        return self.visitChildren(ctx)



del xDroneParser


from __future__ import annotations

import copy


class CollisionConfig:
    def __init__(self, collision_meters: float = 0, time_interval_seconds: float = 0.1,
                 confidence_threshold: float = 1.0):
        if collision_meters < 0:
            raise ValueError("collision_meters should >= 0")
        if time_interval_seconds <= 0:
            raise ValueError("time_interval_seconds should > 0")
        if confidence_threshold < 0 or confidence_threshold > 1:
            raise ValueError("confidence_threshold should between 0 and 1")
        self._collision_meters = collision_meters
        self._time_interval_seconds = time_interval_seconds
        self._confidence_threshold = confidence_threshold

    @property
    def collision_meters(self) -> float:
        return copy.deepcopy(self._collision_meters)

    @property
    def time_interval_seconds(self) -> float:
        return copy.deepcopy(self._time_interval_seconds)

    @property
    def confidence_threshold(self) -> float:
        return copy.deepcopy(self._confidence_threshold)

    def __str__(self):
        return "CollisionConfig: {{ collision_meters: {}, time_interval_seconds: {}, confidence_threshold: {} }}" \
            .format(self._collision_meters, self._time_interval_seconds, self._confidence_threshold)

    def __eq__(self, other):
        if isinstance(other, CollisionConfig):
            return other._collision_meters == self._collision_meters \
                   and other._time_interval_seconds == self._time_interval_seconds \
                   and other._confidence_threshold == self._confidence_threshold


class DefaultCollisionConfig(CollisionConfig):
    def __init__(self):
        super().__init__()



import unittest

from xdrone.command_generators import generate_commands
from xdrone.compiler.compiler_utils.expressions import Expression
from xdrone.compiler.compiler_utils.symbol_table import SymbolTable
from xdrone.compiler.compiler_utils.type import Type
from xdrone.shared.compile_error import CompileError
from xdrone.shared.drone_config import DefaultDroneConfig


class EscapedStringTest(unittest.TestCase):
    def test_normal_string_should_be_parsed_to_correct_expr(self):
        for string in ["", "abc", " ", "abc 123"]:
            actual = SymbolTable()
            generate_commands("""
                main () {{
                  string a <- "{}";
                }}
                """.format(string), symbol_table=actual)
            expected = SymbolTable()
            expected.store("a", Expression(Type.string(), string, ident="a"))
            self.assertEqual(expected, actual)

    def test_escaped_string_should_be_parsed_to_correct_expr(self):
        for escaped, unescaped in [(r"\0", "\0"), (r"\n", "\n"), (r" \"", " \"")]:
            actual = SymbolTable()
            generate_commands("""
                main () {{
                  string a <- "{}";
                }}
                """.format(escaped), symbol_table=actual)
            expected = SymbolTable()
            expected.store("a", Expression(Type.string(), unescaped, ident="a"))
            self.assertEqual(expected, actual)


class BooleanOperationTest(unittest.TestCase):
    def test_bool_or_should_return_correct_value(self):
        for b1, b2 in [("true", "true"), ("true", "false"), ("false", "true"), ("false", "false")]:
            actual = SymbolTable()
            generate_commands("""
                main () {{
                  boolean a <- {} or {};
                }}
                """.format(b1, b2), symbol_table=actual)
            expected = SymbolTable()
            expected.store("a", Expression(Type.boolean(), (b1 == 'true') or (b2 == 'true'), ident="a"))
            self.assertEqual(expected, actual)

    def test_bool_or_with_wrong_type_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      boolean b <- a or true;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type boolean, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      boolean b <- true or a;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type boolean, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_bool_and_should_return_correct_value(self):
        for b1, b2 in [("true", "true"), ("true", "false"), ("false", "true"), ("false", "false")]:
            actual = SymbolTable()
            generate_commands("""
                main () {{
                  boolean a <- {} and {};
                }}
                """.format(b1, b2), symbol_table=actual)
            expected = SymbolTable()
            expected.store("a", Expression(Type.boolean(), (b1 == 'true') and (b2 == 'true'), ident="a"))
            self.assertEqual(expected, actual)

    def test_bool_and_with_wrong_type_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      boolean b <- a and false;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type boolean, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      boolean b <- false and a;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type boolean, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      boolean b <- a or false;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type boolean, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      boolean b <- false or a;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type boolean, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_bool_not_should_return_correct_value(self):
        for b1 in ["true", "false"]:
            actual = SymbolTable()
            generate_commands("""
                main () {{
                  boolean a <- not {};
                }}
                """.format(b1), symbol_table=actual)
            expected = SymbolTable()
            expected.store("a", Expression(Type.boolean(), not (b1 == 'true'), ident="a"))
            self.assertEqual(expected, actual)

    def test_bool_not_with_wrong_type_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      boolean b <- not a;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type boolean, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))


class ComparisonOperationTest(unittest.TestCase):
    def test_greater_should_return_correct_value(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              boolean a <- 1 > -1;
              boolean b <- 1 > -1.0;
              boolean c <- 1.0 > -1;
              boolean d <- 1.0 > -1.0;
              boolean e <- 0 > 0;
              boolean f <- 0.0 > 0.0;
              boolean g <- -1 > 1;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.boolean(), True, ident="a"))
        expected.store("b", Expression(Type.boolean(), True, ident="b"))
        expected.store("c", Expression(Type.boolean(), True, ident="c"))
        expected.store("d", Expression(Type.boolean(), True, ident="d"))
        expected.store("e", Expression(Type.boolean(), False, ident="e"))
        expected.store("f", Expression(Type.boolean(), False, ident="f"))
        expected.store("g", Expression(Type.boolean(), False, ident="g"))
        self.assertEqual(expected, actual)

    def test_greater_with_wrong_type_should_give_error(self):
        types = [Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      boolean b <- a > 1;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      boolean b <- 1 > a;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_greater_equal_should_return_correct_value(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              boolean a <- 1 >= -1;
              boolean b <- 1 >= -1.0;
              boolean c <- 1.0 >= -1;
              boolean d <- 1.0 >= -1.0;
              boolean e <- 0 >= 0;
              boolean f <- 0.0 >= 0.0;
              boolean g <- -1 >= 1;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.boolean(), True, ident="a"))
        expected.store("b", Expression(Type.boolean(), True, ident="b"))
        expected.store("c", Expression(Type.boolean(), True, ident="c"))
        expected.store("d", Expression(Type.boolean(), True, ident="d"))
        expected.store("e", Expression(Type.boolean(), True, ident="e"))
        expected.store("f", Expression(Type.boolean(), True, ident="f"))
        expected.store("g", Expression(Type.boolean(), False, ident="g"))
        self.assertEqual(expected, actual)

    def test_greater_equal_with_wrong_type_should_give_error(self):
        types = [Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      boolean b <- a >= 1;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      boolean b <- 1 >= a;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_less_should_return_correct_value(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              boolean a <- -1 < 1;
              boolean b <- -1 < 1.0;
              boolean c <- -1.0 < 1;
              boolean d <- -1.0 < 1.0;
              boolean e <- 0 < 0;
              boolean f <- 0.0 < 0.0;
              boolean g <- 1 < -1;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.boolean(), True, ident="a"))
        expected.store("b", Expression(Type.boolean(), True, ident="b"))
        expected.store("c", Expression(Type.boolean(), True, ident="c"))
        expected.store("d", Expression(Type.boolean(), True, ident="d"))
        expected.store("e", Expression(Type.boolean(), False, ident="e"))
        expected.store("f", Expression(Type.boolean(), False, ident="f"))
        expected.store("g", Expression(Type.boolean(), False, ident="g"))
        self.assertEqual(expected, actual)

    def test_less_with_wrong_type_should_give_error(self):
        types = [Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      boolean b <- a < 1;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      boolean b <- 1 < a;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_less_equal_should_return_correct_value(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              boolean a <- -1 <= 1;
              boolean b <- -1 <= 1.0;
              boolean c <- -1.0 <= 1;
              boolean d <- -1.0 <= 1.0;
              boolean e <- 0 <= 0;
              boolean f <- 0.0 <= 0.0;
              boolean g <- 1 <= -1;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.boolean(), True, ident="a"))
        expected.store("b", Expression(Type.boolean(), True, ident="b"))
        expected.store("c", Expression(Type.boolean(), True, ident="c"))
        expected.store("d", Expression(Type.boolean(), True, ident="d"))
        expected.store("e", Expression(Type.boolean(), True, ident="e"))
        expected.store("f", Expression(Type.boolean(), True, ident="f"))
        expected.store("g", Expression(Type.boolean(), False, ident="g"))
        self.assertEqual(expected, actual)

    def test_less_equal_with_wrong_type_should_give_error(self):
        types = [Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      boolean b <- a <= 1;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      boolean b <- 1 <= a;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type int or decimal, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_equal_should_return_correct_value(self):
        types = [Type.int(), Type.decimal(), Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            st = SymbolTable()
            generate_commands("""
                main () {{
                  {} a;
                  {} b;
                  boolean c <- a == b;
                }}
                """.format(type, type), symbol_table=st)
            self.assertEqual(Expression(Type.boolean(), True, ident="c"), st.get_expression("c"))

        actual = SymbolTable()
        generate_commands("""
            main () {
              boolean a <- (1.0, 2.0, -3.0) == (1.0, 2.0, -3.0);
              boolean b <- [1, 2] == [1, 2];
              boolean c <- [[1], [2]] == [[1], [2]];
              boolean d <- [] == [];
              boolean e <- "a" == "a";
              boolean f <- DRONE1 == DRONE1;
              
              boolean g <- 1 == 2;
              boolean h <- 1.0 == 2.0;
              boolean i <- true == false;
              boolean j <- "a" == "b";
              boolean k <- (5.0, 2.0, -3.0) == (1.0, 2.0, -3.0);
              boolean l <- [5, 2] == [1, 2];
              boolean n <- [[5], [2]] == [[1], [2]];
              boolean m <- DRONE1 == DRONE2;
            }
            """, drone_config_map={"DRONE1": DefaultDroneConfig(), "DRONE2": DefaultDroneConfig()}, symbol_table=actual)
        expected = SymbolTable()
        for i in "abcdef":
            expected.store(i, Expression(Type.boolean(), True, ident=i))
        for i in "ghijklnm":
            expected.store(i, Expression(Type.boolean(), False, ident=i))
        self.assertEqual(expected, actual)

    def test_equal_diff_types_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for t1 in types:
            for t2 in types:
                if t1 == t2:
                    continue
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        main () {{
                          {} a;
                          {} b;
                          boolean c <- a == b;
                        }}
                        """.format(t1, t2))
                self.assertTrue("Expressions {} and {} should have the same type"
                                .format(Expression(t1, t1.default_value, ident="a"),
                                        Expression(t2, t2.default_value, ident="b"))
                                in str(context.exception))

    def test_not_equal_should_return_correct_value(self):
        types = [Type.int(), Type.decimal(), Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            st = SymbolTable()
            generate_commands("""
                main () {{
                  {} a;
                  {} b;
                  boolean c <- a =/= b;
                }}
                """.format(type, type), symbol_table=st)
            self.assertEqual(Expression(Type.boolean(), False, ident="c"), st.get_expression("c"))

        actual = SymbolTable()
        generate_commands("""
            main () {
              boolean a <- (1.0, 2.0, -3.0) =/= (1.0, 2.0, -3.0);
              boolean b <- [1, 2] =/= [1, 2];
              boolean c <- [[1], [2]] =/= [[1], [2]];
              boolean d <- [] =/= [];
              boolean e <- "a" =/= "a";
              boolean f <- DRONE1 =/= DRONE1;
              
              boolean g <- 1 =/= 2;
              boolean h <- 1.0 =/= 2.0;
              boolean i <- true =/= false;
              boolean j <- "a" =/= "b";
              boolean k <- (5.0, 2.0, -3.0) =/= (1.0, 2.0, -3.0);
              boolean l <- [5, 2] =/= [1, 2];
              boolean n <- [[5], [2]] =/= [[1], [2]];
              boolean m <- DRONE1 =/= DRONE2;
            }
            """, drone_config_map={"DRONE1": DefaultDroneConfig(), "DRONE2": DefaultDroneConfig()}, symbol_table=actual)
        expected = SymbolTable()
        for i in "abcdef":
            expected.store(i, Expression(Type.boolean(), False, ident=i))
        for i in "ghijklnm":
            expected.store(i, Expression(Type.boolean(), True, ident=i))
        self.assertEqual(expected, actual)

    def test_not_equal_diff_types_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.boolean(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for t1 in types:
            for t2 in types:
                if t1 == t2:
                    continue
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        main () {{
                          {} a;
                          {} b;
                          boolean c <- a =/= b;
                        }}
                        """.format(t1, t2))
                self.assertTrue("Expressions {} and {} should have the same type"
                                .format(Expression(t1, t1.default_value, ident="a"),
                                        Expression(t2, t2.default_value, ident="b"))
                                in str(context.exception))


class ArithmeticOperationTest(unittest.TestCase):
    def test_plus_should_return_correct_value(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              int a <- 1 + 1;
              decimal b <- 1 + 1.0;
              decimal c <- 1.0 + 1;
              decimal d <- 1.0 + 1.0;
              vector e <- (1.0, -2.0, 3.0) + (1.0, -2.0, 3.0);
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 2, ident="a"))
        expected.store("b", Expression(Type.decimal(), 2.0, ident="b"))
        expected.store("c", Expression(Type.decimal(), 2.0, ident="c"))
        expected.store("d", Expression(Type.decimal(), 2.0, ident="d"))
        expected.store("e", Expression(Type.vector(), [2.0, -4.0, 6.0], ident="e"))
        self.assertEqual(expected, actual)

    def test_plus_with_wrong_type_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        allow_pairs = [(Type.int(), Type.int()), (Type.decimal(), Type.decimal()), (Type.vector(), Type.vector()),
                       (Type.int(), Type.decimal()), (Type.decimal(), Type.int())]
        for t1 in types:
            for t2 in types:
                if (t1, t2) in allow_pairs:
                    continue
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        main () {{
                          {} a;
                          {} b;
                          int c <- a + b;
                        }}
                        """.format(t1, t2))
                self.assertTrue("Expression {} and {} have wrong types to perform addition or subtraction"
                                .format(Expression(t1, t1.default_value, ident="a"),
                                        Expression(t2, t2.default_value, ident="b"))
                                in str(context.exception))

    def test_minus_should_return_correct_value(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              int a <- 3 - 1;
              decimal b <- 3 - 1.0;
              decimal c <- 3.0 - 1;
              decimal d <- 3.0 - 1.0;
              vector e <- (1.0, -2.0, 3.0) - (-1.0, 2.0, -3.0);
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 2, ident="a"))
        expected.store("b", Expression(Type.decimal(), 2.0, ident="b"))
        expected.store("c", Expression(Type.decimal(), 2.0, ident="c"))
        expected.store("d", Expression(Type.decimal(), 2.0, ident="d"))
        expected.store("e", Expression(Type.vector(), [2.0, -4.0, 6.0], ident="e"))
        self.assertEqual(expected, actual)

    def test_minus_with_wrong_type_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        allow_pairs = [(Type.int(), Type.int()), (Type.decimal(), Type.decimal()), (Type.vector(), Type.vector()),
                       (Type.int(), Type.decimal()), (Type.decimal(), Type.int())]
        for t1 in types:
            for t2 in types:
                if (t1, t2) in allow_pairs:
                    continue
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        main () {{
                          {} a;
                          {} b;
                          int c <- a - b;
                        }}
                        """.format(t1, t2))
                self.assertTrue("Expression {} and {} have wrong types to perform addition or subtraction"
                                .format(Expression(t1, t1.default_value, ident="a"),
                                        Expression(t2, t2.default_value, ident="b"))
                                in str(context.exception))

    def test_multi_should_return_correct_value(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              int a <- 2 * 1;
              decimal b <- 2 * 1.0;
              decimal c <- 2.0 * 1;
              decimal d <- 2.0 * 1.0;
              vector e <- (1.0, 2.0, -3.0) * 2;
              vector f <- (1.0, 2.0, -3.0) * 2.0;
              vector g <- 2 * (1.0, 2.0, -3.0);
              vector h <- 2.0 * (1.0, 2.0, -3.0);
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 2, ident="a"))
        expected.store("b", Expression(Type.decimal(), 2.0, ident="b"))
        expected.store("c", Expression(Type.decimal(), 2.0, ident="c"))
        expected.store("d", Expression(Type.decimal(), 2.0, ident="d"))
        expected.store("e", Expression(Type.vector(), [2.0, 4.0, -6.0], ident="e"))
        expected.store("f", Expression(Type.vector(), [2.0, 4.0, -6.0], ident="f"))
        expected.store("g", Expression(Type.vector(), [2.0, 4.0, -6.0], ident="g"))
        expected.store("h", Expression(Type.vector(), [2.0, 4.0, -6.0], ident="h"))
        self.assertEqual(expected, actual)

    def test_multi_with_wrong_type_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        allow_pairs = [(Type.int(), Type.int()), (Type.decimal(), Type.decimal()),
                       (Type.int(), Type.decimal()), (Type.decimal(), Type.int()),
                       (Type.vector(), Type.int()), (Type.int(), Type.vector()),
                       (Type.vector(), Type.decimal()), (Type.decimal(), Type.vector())]
        for t1 in types:
            for t2 in types:
                if (t1, t2) in allow_pairs:
                    continue
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        main () {{
                          {} a;
                          {} b;
                          int c <- a * b;
                        }}
                        """.format(t1, t2))
                self.assertTrue("Expression {} and {} have wrong types to perform multiplication or division"
                                .format(Expression(t1, t1.default_value, ident="a"),
                                        Expression(t2, t2.default_value, ident="b"))
                                in str(context.exception))

    def test_divide_should_return_correct_value(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              int a <- 2 / 1;
              decimal b <- 2 / 1.0;
              decimal c <- 2.0 / 1;
              decimal d <- 2.0 / 1.0;
              int e <- 1 / 2;
              decimal f <- 1 / 2.0;
              decimal g <- 1.0 / 2;
              decimal h <- 1.0 / 2.0;
              vector i <- (1.0, 2.0, 3.0) / 2;
              vector j <- (1.0, 2.0, 3.0) / 2.0;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 2, ident="a"))
        expected.store("b", Expression(Type.decimal(), 2.0, ident="b"))
        expected.store("c", Expression(Type.decimal(), 2.0, ident="c"))
        expected.store("d", Expression(Type.decimal(), 2.0, ident="d"))
        expected.store("e", Expression(Type.int(), 0, ident="e"))
        expected.store("f", Expression(Type.decimal(), 0.5, ident="f"))
        expected.store("g", Expression(Type.decimal(), 0.5, ident="g"))
        expected.store("h", Expression(Type.decimal(), 0.5, ident="h"))
        expected.store("i", Expression(Type.vector(), [0.5, 1.0, 1.5], ident="i"))
        expected.store("j", Expression(Type.vector(), [0.5, 1.0, 1.5], ident="j"))
        self.assertEqual(expected, actual)

    def test_divide_with_wrong_type_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        allow_pairs = [(Type.int(), Type.int()), (Type.decimal(), Type.decimal()),
                       (Type.int(), Type.decimal()), (Type.decimal(), Type.int()),
                       (Type.vector(), Type.int()),
                       (Type.vector(), Type.decimal())]
        for t1 in types:
            for t2 in types:
                if (t1, t2) in allow_pairs:
                    continue
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        main () {{
                          {} a;
                          {} b;
                          int c <- a / b;
                        }}
                        """.format(t1, t2))
                self.assertTrue("Expression {} and {} have wrong types to perform multiplication or division"
                                .format(Expression(t1, t1.default_value, ident="a"),
                                        Expression(t2, t2.default_value, ident="b"))
                                in str(context.exception))

    def test_divide_by_zero_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {{
                  int a <- 1 / 0;
                }}
                """.format(type))
        self.assertTrue("Division by zero" in str(context.exception))

        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {{
                  decimal a <- 1.0 / 0.0;
                }}
                """.format(type))
        self.assertTrue("Division by zero" in str(context.exception))

    def test_posit_should_return_correct_value(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              int a <- + 1;
              decimal b <- + 1.0;
              vector c <- + (1.0, -1.0, 2.0);
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 1, ident="a"))
        expected.store("b", Expression(Type.decimal(), 1.0, ident="b"))
        expected.store("c", Expression(Type.vector(), [1.0, -1.0, 2.0], ident="c"))
        self.assertEqual(expected, actual)

    def test_posit_with_wrong_type_should_give_error(self):
        types = [Type.boolean(), Type.string(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      {} b <- + a;
                    }}
                    """.format(type, type))
            self.assertTrue("Expression {} should have type int, decimal or vector, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_negate_should_return_correct_value(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              int a <- - 1;
              decimal b <- - 1.0;
              vector c <- - (1.0, -1.0, 2.0);
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), -1, ident="a"))
        expected.store("b", Expression(Type.decimal(), -1.0, ident="b"))
        expected.store("c", Expression(Type.vector(), [-1.0, 1.0, -2.0], ident="c"))
        self.assertEqual(expected, actual)

    def test_negate_with_wrong_type_should_give_error(self):
        types = [Type.boolean(), Type.string(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      {} b <- - a;
                    }}
                    """.format(type, type))
            self.assertTrue("Expression {} should have type int, decimal or vector, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))


class OtherOperationTest(unittest.TestCase):
    def test_size_of_list_should_return_correct_value(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              list[int] a <- [1, 2];
              int b <- a.size;
              int c <- [1.0].size;
              int d <- [].size;
              int e <- [[], [], []].size;
              int f <- [[[], []], [[], []]].size;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.list_of(Type.int()), [1, 2], ident="a"))
        expected.store("b", Expression(Type.int(), 2, ident="b"))
        expected.store("c", Expression(Type.int(), 1, ident="c"))
        expected.store("d", Expression(Type.int(), 0, ident="d"))
        expected.store("e", Expression(Type.int(), 3, ident="e"))
        expected.store("f", Expression(Type.int(), 2, ident="f"))
        self.assertEqual(expected, actual)

    def test_size_of_wrong_type_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.boolean(), Type.string(), Type.vector(), Type.drone()]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      int b <- a.size;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type list, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_concat_of_strings_should_return_correct_value(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              string a <- "a";
              string b <- a & "b";
              string c <- "abc" & "";
              string d <- "" & "abcd";
              string e <- "" & "";
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.string(), "a", ident="a"))
        expected.store("b", Expression(Type.string(), "ab", ident="b"))
        expected.store("c", Expression(Type.string(), "abc", ident="c"))
        expected.store("d", Expression(Type.string(), "abcd", ident="d"))
        expected.store("e", Expression(Type.string(), "", ident="e"))
        self.assertEqual(expected, actual)

    def test_concat_of_wrong_types_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      string b <- a & "";
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type string, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      string b <- "" & a;
                    }}
                    """.format(type))
            self.assertTrue("Expression {} should have type string, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_parentheses_should_visit_content(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              int a <- (1);
              decimal b <- (1.2);
              string c <- ("a");
              boolean d <- (true);
              vector e <- ((1.0, 2.0, (3.0)));
              list[int] f <- ([(1), 2]);
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 1, ident="a"))
        expected.store("b", Expression(Type.decimal(), 1.2, ident="b"))
        expected.store("c", Expression(Type.string(), "a", ident="c"))
        expected.store("d", Expression(Type.boolean(), True, ident="d"))
        expected.store("e", Expression(Type.vector(), [1.0, 2.0, 3.0], ident="e"))
        expected.store("f", Expression(Type.list_of(Type.int()), [1, 2], ident="f"))
        self.assertEqual(expected, actual)


class OperationPrecedenceTest(unittest.TestCase):
    def test_parentheses_higher_than_other(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              int a <- (2 + 3) * 4;
              int b <- (2 + 3) / 4;
              int c <- 1 - (2 - 3);
              boolean d <- not (true and false);
              boolean e <- not (false or true);
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 20, ident="a"))
        expected.store("b", Expression(Type.int(), 1, ident="b"))
        expected.store("c", Expression(Type.int(), 2, ident="c"))
        expected.store("d", Expression(Type.boolean(), True, ident="d"))
        expected.store("e", Expression(Type.boolean(), False, ident="e"))
        self.assertEqual(expected, actual)

    def test_negate_higher_than_plus_minus(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              int a <- - 2 + 3;
              int b <- - 2 - 3;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 1, ident="a"))
        expected.store("b", Expression(Type.int(), -5, ident="b"))
        self.assertEqual(expected, actual)

    def test_multi_divide_same_precedence(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              int a <- 2 * 2 / 3;
              int b <- 2 / 2 * 3;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 1, ident="a"))
        expected.store("b", Expression(Type.int(), 3, ident="b"))
        self.assertEqual(expected, actual)

    def test_plus_minus_same_precedence(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              int a <- 1 - 2 + 3;
              int b <- 1 + 2 - 3;
              int c <- 1 - 2 - 3;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 2, ident="a"))
        expected.store("b", Expression(Type.int(), 0, ident="b"))
        expected.store("c", Expression(Type.int(), -4, ident="c"))
        self.assertEqual(expected, actual)

    def test_multi_divide_higher_than_plus_minus(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              int a <- 2 + 3 * 4;
              int b <- 2 - 3 * 4;
              int c <- 2 + 3 / 4;
              int d <- 2 - 3 / 4;
              int e <- 3 * 4 + 2;
              int f <- 3 * 4 - 2;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 14, ident="a"))
        expected.store("b", Expression(Type.int(), -10, ident="b"))
        expected.store("c", Expression(Type.int(), 2, ident="c"))
        expected.store("d", Expression(Type.int(), 2, ident="d"))
        expected.store("e", Expression(Type.int(), 14, ident="e"))
        expected.store("f", Expression(Type.int(), 10, ident="f"))
        self.assertEqual(expected, actual)

    def test_arithmetic_higher_than_comparison(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              boolean a <- 2 + 3 == 5;
              boolean b <- 2 * 3 =/= 5;
              boolean c <- 2 - 3 <= 5;
              boolean d <- 2 - 3 < 5;
              boolean e <- 2 / 3 >= 5;
              boolean f <- 2 / 3 > 5;
              boolean g <- 5 == 2 + 3;
              boolean h <- 5 =/= 2 * 3;
              boolean i <- 5 <= 2 - 3;
              boolean j <- 5 < 2 - 3;
              boolean k <- 5 >= 2 / 3;
              boolean l <- 5 > 2 / 3;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.boolean(), True, ident="a"))
        expected.store("b", Expression(Type.boolean(), True, ident="b"))
        expected.store("c", Expression(Type.boolean(), True, ident="c"))
        expected.store("d", Expression(Type.boolean(), True, ident="d"))
        expected.store("e", Expression(Type.boolean(), False, ident="e"))
        expected.store("f", Expression(Type.boolean(), False, ident="f"))
        expected.store("g", Expression(Type.boolean(), True, ident="g"))
        expected.store("h", Expression(Type.boolean(), True, ident="h"))
        expected.store("i", Expression(Type.boolean(), False, ident="i"))
        expected.store("j", Expression(Type.boolean(), False, ident="j"))
        expected.store("k", Expression(Type.boolean(), True, ident="k"))
        expected.store("l", Expression(Type.boolean(), True, ident="l"))
        self.assertEqual(expected, actual)

    def test_comparison_higher_than_equality(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              boolean a <- true == 6 > 5;
              boolean b <- true == 6 >= 5;
              boolean c <- true == 6 < 5;
              boolean d <- true == 6 <= 5;
              boolean e <- 6 < 5 =/= true;
              boolean f <- 6 <= 5 =/= true;
              boolean g <- 6 > 5 =/= true;
              boolean h <- 6 >= 5 =/= true;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.boolean(), True, ident="a"))
        expected.store("b", Expression(Type.boolean(), True, ident="b"))
        expected.store("c", Expression(Type.boolean(), False, ident="c"))
        expected.store("d", Expression(Type.boolean(), False, ident="d"))
        expected.store("e", Expression(Type.boolean(), True, ident="e"))
        expected.store("f", Expression(Type.boolean(), True, ident="f"))
        expected.store("g", Expression(Type.boolean(), False, ident="g"))
        expected.store("h", Expression(Type.boolean(), False, ident="h"))
        self.assertEqual(expected, actual)

    def test_equality_same_precedence(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              boolean a <- 2 + 3 == 5 == true;
              boolean b <- 2 * 3 =/= 5 =/= false;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.boolean(), True, ident="a"))
        expected.store("b", Expression(Type.boolean(), True, ident="b"))
        self.assertEqual(expected, actual)

    def test_comparison_higher_than_bool_operation(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              boolean a <- true and 1 == 1;
              boolean b <- false or 1 == 1;
              boolean c <- 1 == 1 and true;
              boolean d <- 1 == 1 or false;
              boolean e <- true and 1 >= 1;
              boolean f <- false or 1 <= 1;
              boolean g <- 1 > 1 and true;
              boolean h <- 1 < 1 or false;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.boolean(), True, ident="a"))
        expected.store("b", Expression(Type.boolean(), True, ident="b"))
        expected.store("c", Expression(Type.boolean(), True, ident="c"))
        expected.store("d", Expression(Type.boolean(), True, ident="d"))
        expected.store("e", Expression(Type.boolean(), True, ident="e"))
        expected.store("f", Expression(Type.boolean(), True, ident="f"))
        expected.store("g", Expression(Type.boolean(), False, ident="g"))
        expected.store("h", Expression(Type.boolean(), False, ident="h"))
        self.assertEqual(expected, actual)

    def test_concat_higher_than_eq_operation(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              boolean a <- "a" & "b" == "ab";
              boolean b <- "ab" == "a" & "b";
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.boolean(), True, ident="a"))
        expected.store("b", Expression(Type.boolean(), True, ident="b"))
        self.assertEqual(expected, actual)

    def test_not_higher_than_and_or(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              boolean a <- not true and false;
              boolean b <- not false or true;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.boolean(), False, ident="a"))
        expected.store("b", Expression(Type.boolean(), True, ident="b"))
        self.assertEqual(expected, actual)


class OperationConfusionTest(unittest.TestCase):
    def test_posit_not_confused_with_plus(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              int a <- 1 + 2;
              int b <- 1+2;
              int c <- +2;
              int d <- 1++2;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 3, ident="a"))
        expected.store("b", Expression(Type.int(), 3, ident="b"))
        expected.store("c", Expression(Type.int(), 2, ident="c"))
        expected.store("d", Expression(Type.int(), 3, ident="d"))
        self.assertEqual(expected, actual)

    def test_negate_not_confused_with_minus(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
              int a <- 1 - 2;
              int b <- 1-2;
              int c <- -2;
              int d <- 1--2;
              int e <- 1---2;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), -1, ident="a"))
        expected.store("b", Expression(Type.int(), -1, ident="b"))
        expected.store("c", Expression(Type.int(), -2, ident="c"))
        expected.store("d", Expression(Type.int(), 3, ident="d"))
        expected.store("e", Expression(Type.int(), -1, ident="e"))
        self.assertEqual(expected, actual)



import json
from logging import warning
from typing import Dict

from xdrone.shared.boundary_config import BoundaryConfig
from xdrone.shared.collision_config import CollisionConfig, DefaultCollisionConfig
from xdrone.shared.drone_config import DroneConfig, DefaultDroneConfig


class ConfigParser:
    @staticmethod
    def parse(json_str: str) -> (Dict[str, DroneConfig], BoundaryConfig, CollisionConfig):
        data = json.loads(json_str)
        drone_config_map = ConfigParser._parse_drone_config(data)
        boundary_config = ConfigParser._parse_boundary_config(data)
        collision_config = ConfigParser._parse_collision_config(data)
        return drone_config_map, boundary_config, collision_config

    @staticmethod
    def _parse_drone_config(data: dict) -> Dict[str, DroneConfig]:
        drone_config_map = {}
        if "drones" in data:
            if len(data["drones"]) == 0:
                warning("'drones' is empty when parsing configs, using default drone_config. " +
                        "Position estimation may be inaccurate.")
                return {"DEFAULT": DefaultDroneConfig()}
            for drone in data["drones"]:
                if "name" in drone:
                    name = drone["name"]
                    if name == "":
                        warning("'name' cannot be an empty string, using default value 'DEFAULT' instead.")
                        name = "DEFAULT"
                else:
                    warning("'name' missing when parsing object in 'drones', using default value 'DEFAULT'.")
                    name = "DEFAULT"

                if name in drone_config_map:
                    warning("Drone name '{}' appeared more than ones in 'drones', ignored.".format(name))
                    continue

                if "init_position" in drone:
                    position = drone["init_position"]
                    init_position = []
                    for dim in ["x", "y", "z"]:
                        if dim in position:
                            init_position.append(position[dim])
                        else:
                            warning("'{}' missing when parsing drone '{}', using default value 0. ".format(dim, name) +
                                    "Position estimation may be inaccurate.")
                            init_position.append(0)
                    init_position = tuple(init_position)
                else:
                    warning("'init_position' missing when parsing drone '{}', ".format(name) +
                            "using default value (0, 0, 0). " +
                            "Position estimation may be inaccurate.")
                    init_position = (0, 0, 0)
                if "speed_mps" in drone:
                    speed_mps = drone["speed_mps"]
                else:
                    warning("'speed_mps' missing when parsing drone '{}', ".format(name) +
                            "using default value 1. " +
                            "Position estimation may be inaccurate.")
                    speed_mps = 1
                if "rotate_speed_dps" in drone:
                    rotate_speed_dps = drone["rotate_speed_dps"]
                else:
                    warning("'rotate_speed_dps' missing when parsing drone '{}', ".format(name) +
                            "using default value 90. " +
                            "Position estimation may be inaccurate.")
                    rotate_speed_dps = 90
                if "takeoff_height_meters" in drone:
                    takeoff_height_meters = drone["takeoff_height_meters"]
                else:
                    warning("'takeoff_height_meters' missing when parsing drone '{}', ".format(name) +
                            "using default value 1. " +
                            "Position estimation may be inaccurate.")
                    takeoff_height_meters = 1

                # no warning given if fields in advanced missing
                var_per_meter = 0.0
                if "advanced" in drone:
                    if "variance_per_meter" in drone["advanced"]:
                        var_per_meter = drone["advanced"]["variance_per_meter"]

                # TODO: variance in position when rotation is confused, better to change to variance in orientation
                variance_per_degree = 0.0

                drone_config = DroneConfig(init_position, speed_mps, rotate_speed_dps, takeoff_height_meters,
                                           var_per_meter, variance_per_degree)
                drone_config_map[name] = drone_config
        else:
            warning("'drones' missing when parsing configs, using default drone_config. " +
                    "Position estimation may be inaccurate.")
            return {"DEFAULT": DefaultDroneConfig()}
        return drone_config_map

    @staticmethod
    def _parse_boundary_config(data: dict) -> BoundaryConfig:
        if "boundary_config" in data:
            if "max_seconds" in data["boundary_config"]:
                max_seconds = data["boundary_config"]["max_seconds"]
            else:
                warning("'max_seconds' missing when parsing 'boundary_config', using default value inf. " +
                        "There will be no limit on 'max_seconds'.")
                max_seconds = float('inf')
            if "max_x_meters" in data["boundary_config"]:
                max_x_meters = data["boundary_config"]["max_x_meters"]
            else:
                warning("'max_x_meters' missing when parsing 'boundary_config', using default value inf. " +
                        "There will be no limit on 'max_x_meters'.")
                max_x_meters = float('inf')
            if "max_y_meters" in data["boundary_config"]:
                max_y_meters = data["boundary_config"]["max_y_meters"]
            else:
                warning("'max_y_meters' missing when parsing 'boundary_config', using default value inf. " +
                        "There will be no limit on 'max_y_meters'.")
                max_y_meters = float('inf')
            if "max_z_meters" in data["boundary_config"]:
                max_z_meters = data["boundary_config"]["max_z_meters"]
            else:
                warning("'max_z_meters' missing when parsing 'boundary_config', using default value inf. " +
                        "There will be no limit on 'max_z_meters'.")
                max_z_meters = float('inf')
            if "min_x_meters" in data["boundary_config"]:
                min_x_meters = data["boundary_config"]["min_x_meters"]
            else:
                warning("'min_x_meters' missing when parsing 'boundary_config', using default value -inf. " +
                        "There will be no limit on 'min_x_meters'.")
                min_x_meters = float('-inf')
            if "min_y_meters" in data["boundary_config"]:
                min_y_meters = data["boundary_config"]["min_y_meters"]
            else:
                warning("'min_y_meters' missing when parsing 'boundary_config', using default value -inf. " +
                        "There will be no limit on 'min_y_meters'.")
                min_y_meters = float('-inf')
            if "min_z_meters" in data["boundary_config"]:
                min_z_meters = data["boundary_config"]["min_z_meters"]
            else:
                warning("'min_z_meters' missing when parsing 'boundary_config', using default value -inf. " +
                        "There will be no limit on 'min_z_meters'.")
                min_z_meters = float('-inf')
            boundary_config = BoundaryConfig(max_seconds, max_x_meters, max_y_meters, max_z_meters,
                                             min_x_meters, min_y_meters, min_z_meters)
        else:
            warning("'boundary_config' missing when parsing configs, using unlimited boundary_config. " +
                    "Time and position will be unlimited.")
            boundary_config = BoundaryConfig.no_limit()
        return boundary_config

    @staticmethod
    def _parse_collision_config(data: dict) -> CollisionConfig:
        if "collision_config" in data:
            if "collision_meters" in data["collision_config"]:
                collision_meters = data["collision_config"]["collision_meters"]
            else:
                warning("'collision_meters' missing when parsing 'collision_config', using default value 0. " +
                        "There will be no limit on 'collision_meters'.")
                collision_meters = 0.0
            if "time_interval_seconds" in data["collision_config"]:
                collision_time_interval_seconds = data["collision_config"]["time_interval_seconds"]
            else:
                warning("'time_interval_seconds' missing when parsing 'collision_config', " +
                        "using default value 0.1.")
                collision_time_interval_seconds = 0.1
            if "confidence_threshold" in data["collision_config"]:
                confidence_threshold = data["collision_config"]["confidence_threshold"]
            else:
                warning("'confidence_threshold' missing when parsing 'collision_config', " +
                        "using default value 1.0.")
                confidence_threshold = 1.0
            collision_config = CollisionConfig(collision_meters, collision_time_interval_seconds, confidence_threshold)
        else:
            warning("'collision_config' missing when parsing configs, using default collision_config.")
            collision_config = DefaultCollisionConfig()
        return collision_config



import json
import unittest
import zlib
from base64 import b64decode

from xdrone.command_converters.simulation_converter import SimulationConverter
from xdrone.shared.command import Command, SingleDroneCommand, ParallelDroneCommands
from xdrone.shared.drone_config import DroneConfig


class TestSimulationConverter(unittest.TestCase):
    def test_convert_command(self):
        commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                    SingleDroneCommand("DRONE1", Command.up(1)),
                    SingleDroneCommand("DRONE1", Command.down(1)),
                    SingleDroneCommand("DRONE1", Command.left(1)),
                    SingleDroneCommand("DRONE1", Command.right(1)),
                    SingleDroneCommand("DRONE1", Command.forward(1)),
                    SingleDroneCommand("DRONE1", Command.backward(1)),
                    SingleDroneCommand("DRONE1", Command.rotate_left(1)),
                    SingleDroneCommand("DRONE1", Command.rotate_right(1)),
                    ParallelDroneCommands([
                        [ParallelDroneCommands([
                            [],
                            [SingleDroneCommand("DRONE1", Command.up(1))]
                        ])],
                        [SingleDroneCommand("DRONE2", Command.takeoff()), SingleDroneCommand("DRONE2", Command.land())]

                    ]),
                    SingleDroneCommand("DRONE1", Command.land()),
                    SingleDroneCommand("DRONE1", Command.wait(1))]
        drone_config_map = {"DRONE1": DroneConfig((1, 2, 3), 1, 90, 2), "DRONE2": DroneConfig((0, 0, 0), 1, 90, 1)}
        expected = {"config": [{"name": "DRONE1", "init_pos": [1, 2, 3],
                                "speed": 1, "rotate_speed": 90, "takeoff_height": 2},
                               {"name": "DRONE2", "init_pos": [0, 0, 0],
                                "speed": 1, "rotate_speed": 90, "takeoff_height": 1}],
                    "commands": [{"type": "single", "name": "DRONE1", "action": "takeoff", "values": []},
                                 {"type": "single", "name": "DRONE1", "action": "up", "values": [1]},
                                 {"type": "single", "name": "DRONE1", "action": "down", "values": [1]},
                                 {"type": "single", "name": "DRONE1", "action": "left", "values": [1]},
                                 {"type": "single", "name": "DRONE1", "action": "right", "values": [1]},
                                 {"type": "single", "name": "DRONE1", "action": "forward", "values": [1]},
                                 {"type": "single", "name": "DRONE1", "action": "backward", "values": [1]},
                                 {"type": "single", "name": "DRONE1", "action": "rotate_left", "values": [1]},
                                 {"type": "single", "name": "DRONE1", "action": "rotate_right", "values": [1]},
                                 {"type": "parallel", "branches": [
                                     [{"type": "parallel", "branches": [
                                         [],
                                         [{"type": "single", "name": "DRONE1", "action": "up", "values": [1]}]]}],
                                     [{"type": "single", "name": "DRONE2", "action": "takeoff", "values": []},
                                      {"type": "single", "name": "DRONE2", "action": "land", "values": []}]
                                 ]},
                                 {"type": "single", "name": "DRONE1", "action": "land", "values": []},
                                 {"type": "single", "name": "DRONE1", "action": "wait", "values": [1]}
                                 ]}
        actual = SimulationConverter().convert(commands, drone_config_map)

        self.assertEqual(expected, json.loads(zlib.decompress(b64decode(actual))))



import unittest
from unittest.mock import Mock, call, patch

from xdrone.command_converters.dji_tello_edu_drone_executor import DJITelloEduExecutor
from xdrone.shared.command import SingleDroneCommand, Command, ParallelDroneCommands


class TestDJITelloEduExecutor(unittest.TestCase):
    def setUp(self):
        self.fly = Mock()
        self.fly.attach_mock(Mock(), "wait_sync")
        self.fly.attach_mock(Mock(), "takeoff")
        self.fly.attach_mock(Mock(), "land")
        self.fly.attach_mock(Mock(), "up")
        self.fly.attach_mock(Mock(), "down")
        self.fly.attach_mock(Mock(), "left")
        self.fly.attach_mock(Mock(), "right")
        self.fly.attach_mock(Mock(), "forward")
        self.fly.attach_mock(Mock(), "back")
        self.fly.attach_mock(Mock(), "rotate_ccw")
        self.fly.attach_mock(Mock(), "rotate_cw")
        self.fly.attach_mock(Mock(), "pause")

        self.fly_tello = Mock()
        self.fly_tello.__enter__ = Mock(return_value=self.fly)
        self.fly_tello.__exit__ = Mock()

        self.name_id_map = {"name1": "id1", "name2": "id2"}

    def test_execute_single_drone_command(self):
        single_commands = [SingleDroneCommand("name1", Command.takeoff()),
                           SingleDroneCommand("name1", Command.land()),
                           SingleDroneCommand("name2", Command.up(1.1)),
                           SingleDroneCommand("name2", Command.down(1.2)),
                           SingleDroneCommand("name2", Command.left(1.3)),
                           SingleDroneCommand("name2", Command.right(1.4)),
                           SingleDroneCommand("name2", Command.forward(1.5)),
                           SingleDroneCommand("name2", Command.backward(1.6)),
                           SingleDroneCommand("name2", Command.rotate_left(91)),
                           SingleDroneCommand("name2", Command.rotate_right(92)),
                           SingleDroneCommand("name2", Command.wait(1))]
        with patch('xdrone.command_converters.dji_tello_edu_drone_executor.FlyTello', return_value=self.fly_tello):
            executor = DJITelloEduExecutor(self.name_id_map)
            executor.execute_drone_commands(single_commands)
            calls = [call.wait_sync(),
                     call.takeoff(tello=1),
                     call.wait_sync(),
                     call.land(tello=1),
                     call.wait_sync(),
                     call.up(110, tello=2),
                     call.wait_sync(),
                     call.down(120, tello=2),
                     call.wait_sync(),
                     call.left(130, tello=2),
                     call.wait_sync(),
                     call.right(140, tello=2),
                     call.wait_sync(),
                     call.forward(150, tello=2),
                     call.wait_sync(),
                     call.back(160, tello=2),
                     call.wait_sync(),
                     call.rotate_ccw(91, tello=2),
                     call.wait_sync(),
                     call.rotate_cw(92, tello=2),
                     call.wait_sync(),
                     call.pause(1)]
            self.fly.assert_has_calls(calls)

    def test_execute_parallel_drone_commands(self):
        parallel_commands = [
            ParallelDroneCommands([
                [SingleDroneCommand("name1", Command.takeoff()), SingleDroneCommand("name1", Command.land())],
                [SingleDroneCommand("name2", Command.takeoff()), SingleDroneCommand("name2", Command.land())]
            ])
        ]
        with patch('xdrone.command_converters.dji_tello_edu_drone_executor.FlyTello', return_value=self.fly_tello):
            executor = DJITelloEduExecutor(self.name_id_map)
            executor.execute_drone_commands(parallel_commands)
            calls = [call.wait_sync(),
                     call.takeoff(tello=1),
                     call.land(tello=1),
                     call.takeoff(tello=2),
                     call.land(tello=2)]
            self.fly.assert_has_calls(calls)

    def test_execute_nested_parallel_drone_commands(self):
        nested_parallel_commands = [
            ParallelDroneCommands([
                [ParallelDroneCommands([
                    [SingleDroneCommand("name1", Command.takeoff()), SingleDroneCommand("name1", Command.land())]
                ])],
                [SingleDroneCommand("name2", Command.takeoff()), SingleDroneCommand("name2", Command.land())]
            ])
        ]
        with patch('xdrone.command_converters.dji_tello_edu_drone_executor.FlyTello', return_value=self.fly_tello):
            executor = DJITelloEduExecutor(self.name_id_map)
            executor.execute_drone_commands(nested_parallel_commands)
            calls = [call.wait_sync(),
                     call.takeoff(tello=1),
                     call.land(tello=1),
                     call.takeoff(tello=2),
                     call.land(tello=2)]
            self.fly.assert_has_calls(calls)

    def test_execute_mixed_drone_commands(self):
        mixed_commands = [
            SingleDroneCommand("name1", Command.takeoff()),
            ParallelDroneCommands([
                [SingleDroneCommand("name1", Command.land())],
                [SingleDroneCommand("name2", Command.takeoff())]
            ]),
            SingleDroneCommand("name2", Command.land())
        ]
        with patch('xdrone.command_converters.dji_tello_edu_drone_executor.FlyTello', return_value=self.fly_tello):
            executor = DJITelloEduExecutor(self.name_id_map)
            executor.execute_drone_commands(mixed_commands)
            calls = [call.wait_sync(),
                     call.takeoff(tello=1),
                     call.wait_sync(),
                     call.land(tello=1),
                     call.takeoff(tello=2),
                     call.wait_sync(),
                     call.land(tello=2)]
            self.fly.assert_has_calls(calls)
